import pytest

import salt.utils.files
import salt.utils.win_lgpo_reg as win_lgpo_reg
from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


def test_dict_to_reg_pol_empty():
    """
    Test an empty dict. Should only contain the header
    """
    test_dict = {}
    expected = b"PReg\x01\x00\x00\x00"
    result = win_lgpo_reg.dict_to_reg_pol(test_dict)
    assert result == expected


def test_reg_pol_to_dict_empty():
    """
    Test reg pol data with nothing configured. Only the header. Should return
    an empty dict
    """
    test_data = b"PReg\x01\x00\x00\x00"
    expected = {}
    result = win_lgpo_reg.reg_pol_to_dict(test_data)
    assert result == expected


def test_reg_pol_to_dict_empty_invalid():
    """
    Test reg pol data with nothing configured. Only the header. Should return
    an empty dict
    """
    test_data = b""
    pytest.raises(CommandExecutionError, win_lgpo_reg.reg_pol_to_dict, test_data)


def test_dict_to_reg_pol_reg_sz():
    """
    Test REG_SZ type
    """
    test_dict = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": "String",
                "type": "REG_SZ",
            },
        },
    }
    expected = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x01\x00\x00\x00;\x00"
        b"\x0e\x00\x00\x00;\x00"
        b"S\x00t\x00r\x00i\x00n\x00g\x00\x00\x00"
        b"]\x00"
    )
    result = win_lgpo_reg.dict_to_reg_pol(test_dict)
    assert result == expected


def test_reg_pol_to_dict_reg_sz():
    """
    Test REG_SZ type
    """
    test_data = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x01\x00\x00\x00;\x00"
        b"\x0e\x00\x00\x00;\x00"
        b"S\x00t\x00r\x00i\x00n\x00g\x00\x00\x00"
        b"]\x00"
    )
    expected = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": "String",
                "type": "REG_SZ",
            },
        },
    }
    result = win_lgpo_reg.reg_pol_to_dict(test_data)
    assert result == expected


def test_dict_to_reg_pol_reg_expand_sz():
    """
    Test REG_EXPAND_SZ type
    """
    test_dict = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": "%WINVER%\\String",
                "type": "REG_EXPAND_SZ",
            },
        },
    }
    expected = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x02\x00\x00\x00;\x00"
        b" \x00\x00\x00;\x00"
        b"%\x00W\x00I\x00N\x00V\x00E\x00R\x00%\x00\\\x00S\x00t\x00r\x00i\x00n\x00g\x00\x00\x00"
        b"]\x00"
    )
    result = win_lgpo_reg.dict_to_reg_pol(test_dict)
    assert result == expected


def test_reg_pol_to_dict_reg_expand_sz():
    """
    Test REG_EXPAND_SZ type
    """
    test_data = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x02\x00\x00\x00;\x00"
        b" \x00\x00\x00;\x00"
        b"%\x00W\x00I\x00N\x00V\x00E\x00R\x00%\x00\\\x00S\x00t\x00r\x00i\x00n\x00g\x00\x00\x00"
        b"]\x00"
    )
    expected = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": "%WINVER%\\String",
                "type": "REG_EXPAND_SZ",
            },
        },
    }
    result = win_lgpo_reg.reg_pol_to_dict(test_data)
    assert result == expected


def test_dict_to_reg_pol_reg_dword():
    """
    Test REG_DWORD type
    """
    test_dict = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": 1,
                "type": "REG_DWORD",
            },
        },
    }
    expected = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x04\x00\x00\x00;\x00"
        b"\x04\x00\x00\x00;\x00"
        b"\x01\x00\x00\x00"
        b"]\x00"
    )
    result = win_lgpo_reg.dict_to_reg_pol(test_dict)
    assert result == expected


def test_reg_pol_to_dict_reg_dword():
    """
    Test REG_DWORD type
    """
    test_data = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x04\x00\x00\x00;\x00"
        b"\x04\x00\x00\x00;\x00"
        b"\x01\x00\x00\x00"
        b"]\x00"
    )
    expected = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": 1,
                "type": "REG_DWORD",
            },
        },
    }
    result = win_lgpo_reg.reg_pol_to_dict(test_data)
    assert result == expected


def test_dict_to_reg_pol_reg_multi_sz():
    """
    Test REG_MULTI_SZ type
    """
    test_dict = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": ["rick", "morty"],
                "type": "REG_MULTI_SZ",
            },
        },
    }
    expected = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x07\x00\x00\x00;\x00"
        b"\x18\x00\x00\x00;\x00"
        b"r\x00i\x00c\x00k\x00\x00\x00m\x00o\x00r\x00t\x00y\x00\x00\x00\x00\x00"
        b"]\x00"
    )
    result = win_lgpo_reg.dict_to_reg_pol(test_dict)
    assert result == expected


def test_reg_pol_to_dict_reg_multi_sz():
    """
    Test REG_MULTI_SZ type
    """
    test_data = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x07\x00\x00\x00;\x00"
        b"\x18\x00\x00\x00;\x00"
        b"r\x00i\x00c\x00k\x00\x00\x00m\x00o\x00r\x00t\x00y\x00\x00\x00\x00\x00"
        b"]\x00"
    )
    expected = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": ["rick", "morty"],
                "type": "REG_MULTI_SZ",
            },
        },
    }
    result = win_lgpo_reg.reg_pol_to_dict(test_data)
    assert result == expected


def test_dict_to_reg_pol_reg_multi_sz_none():
    """
    Test REG_MULTI_SZ type when the value is None
    """
    test_dict = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": None,
                "type": "REG_MULTI_SZ",
            },
        },
    }
    expected = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x07\x00\x00\x00;\x00"
        b"\x02\x00\x00\x00;\x00"
        b"\x00\x00"
        b"]\x00"
    )
    result = win_lgpo_reg.dict_to_reg_pol(test_dict)
    assert result == expected


def test_dict_to_reg_pol_reg_multi_sz_empty_list():
    """
    Test REG_MULTI_SZ type when the value is an empty list
    """
    test_dict = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": [],
                "type": "REG_MULTI_SZ",
            },
        },
    }
    expected = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x07\x00\x00\x00;\x00"
        b"\x02\x00\x00\x00;\x00"
        b"\x00\x00"
        b"]\x00"
    )
    result = win_lgpo_reg.dict_to_reg_pol(test_dict)
    assert result == expected


def test_dict_to_reg_pol_reg_multi_sz_empty_list_value():
    """
    Test REG_MULTI_SZ type when the value is a list with an empty value
    """
    test_dict = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": [""],
                "type": "REG_MULTI_SZ",
            },
        },
    }
    expected = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x07\x00\x00\x00;\x00"
        b"\x02\x00\x00\x00;\x00"
        b"\x00\x00"
        b"]\x00"
    )
    result = win_lgpo_reg.dict_to_reg_pol(test_dict)
    assert result == expected


def test_dict_to_reg_pol_reg_multi_sz_list_single_value():
    """
    Test REG_MULTI_SZ type when the value is a list with an empty value
    """
    test_dict = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": ["rick"],
                "type": "REG_MULTI_SZ",
            },
        },
    }
    expected = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x07\x00\x00\x00;\x00"
        b"\x0c\x00\x00\x00;\x00"
        b"r\x00i\x00c\x00k\x00\x00\x00\x00\x00"
        b"]\x00"
    )
    result = win_lgpo_reg.dict_to_reg_pol(test_dict)
    assert result == expected


def test_reg_pol_to_dict_reg_multi_sz_empty():
    """
    Test REG_MULTI_SZ type when the value is empty
    """
    test_data = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x07\x00\x00\x00;\x00"
        b"\x02\x00\x00\x00;\x00"
        b"\x00\x00"
        b"]\x00"
    )
    expected = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": None,
                "type": "REG_MULTI_SZ",
            },
        },
    }
    result = win_lgpo_reg.reg_pol_to_dict(test_data)
    assert result == expected


def test_dict_to_reg_pol_reg_qword():
    """
    Test REG_QWORD
    """
    test_dict = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": 1,
                "type": "REG_QWORD",
            },
        },
    }
    expected = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x0b\x00\x00\x00;\x00"
        b"\x08\x00\x00\x00;\x00"
        b"\x01\x00\x00\x00\x00\x00\x00\x00"
        b"]\x00"
    )
    result = win_lgpo_reg.dict_to_reg_pol(test_dict)
    assert result == expected


def test_reg_pol_to_dict_reg_qword():
    """
    Test REG_QWORD
    """
    test_data = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x0b\x00\x00\x00;\x00"
        b"\x08\x00\x00\x00;\x00"
        b"\x01\x00\x00\x00\x00\x00\x00\x00"
        b"]\x00"
    )
    expected = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": 1,
                "type": "REG_QWORD",
            },
        },
    }
    result = win_lgpo_reg.reg_pol_to_dict(test_data)
    assert result == expected


def test_reg_pol_to_dict_invalid_type():
    """
    Test Invalid Registry Type
    """
    test_data = (
        b"PReg\x01\x00\x00\x00[\x00"
        b"S\x00O\x00F\x00T\x00W\x00A\x00R\x00E\x00\\\x00M\x00y\x00K\x00e\x00y\x00\x00\x00;\x00"
        b"M\x00y\x00V\x00a\x00l\x00u\x00e\x00\x00\x00;\x00"
        b"\x1b\x00\x00\x00;\x00"
        b"\x04\x00\x00\x00;\x00"
        b"\x01\x00\x00\x00"
        b"]\x00"
    )
    pytest.raises(CommandExecutionError, win_lgpo_reg.reg_pol_to_dict, test_data)


def test_dict_to_reg_pol_invalid_type():
    """
    Test Invalid Registry Type
    """
    test_dict = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": 1,
                "type": "REG_INVALID_TYPE",
            },
        },
    }
    pytest.raises(CommandExecutionError, win_lgpo_reg.dict_to_reg_pol, test_dict)


def test_dict_to_reg_pol_too_big():
    """
    Test when the data exceeds 65535 characters
    """
    test_dict = {
        "SOFTWARE\\MyKey": {
            "MyValue": {
                "data": "*" * 65536,
                "type": "REG_SZ",
            },
        },
    }
    pytest.raises(CommandExecutionError, win_lgpo_reg.dict_to_reg_pol, test_dict)


def test_issue_56769_windows_line_endings():
    """
    Test that it handles a gpt.ini file with Windows-style line endings.
    Should create a gpt.ini with Windows-style line endings.
    """

    data_to_write = b"[\x00d\x00u\x00m\x00m\x00y\x00\\\x00d\x00a\x00t\x00a]\x00"
    gpt_extension = "gPCMachineExtensionNames"
    gpt_extension_guid = (
        "[{35378EAC-683F-11D2-A89A-00C04FBBCFA2}{D02B1F72-3407-48AE-BA88-E8213C6761F1}]"
    )

    gpt_ini = "\r\n".join(["[General]", "gPCMachineExtensionNames=", "Version=8", ""])
    expected = "\r\n".join(
        [
            "[General]",
            "gPCMachineExtensionNames=[{35378EAC-683F-11D2-A89A-00C04FBBCFA2}{D02B1F72-3407-48AE-BA88-E8213C6761F1}]",
            "Version=9",
            "",
        ]
    )

    with pytest.helpers.temp_file(
        "Registry.pol"
    ) as reg_pol_file, pytest.helpers.temp_file("gpt.ini") as gpt_ini_file:
        # We're using salt.utils.file.fopen here because the temp_file helper
        # doesn't preserve line endings when writing the test file
        with salt.utils.files.fopen(str(gpt_ini_file), "wb") as fp:
            fp.write(gpt_ini.encode("utf-8"))
        win_lgpo_reg.write_reg_pol_data(
            data_to_write=data_to_write,
            policy_file_path=str(reg_pol_file),
            gpt_extension=gpt_extension,
            gpt_extension_guid=gpt_extension_guid,
            gpt_ini_path=str(gpt_ini_file),
        )
        # We're using salt.utils.file.fopen here because the temp_file helper
        # doesn't preserve line endings when reading the test file
        with salt.utils.files.fopen(str(gpt_ini_file)) as fp:
            result = fp.read()

        assert result == expected


def test_issue_56769_unix_line_endings():
    """
    Test that it handles a gpt.ini file with Unix-style line endings.
    Should create a gpt.ini with Windows-style line endings.
    """

    data_to_write = b"[\x00d\x00u\x00m\x00m\x00y\x00\\\x00d\x00a\x00t\x00a]\x00"
    gpt_extension = "gPCMachineExtensionNames"
    gpt_extension_guid = (
        "[{35378EAC-683F-11D2-A89A-00C04FBBCFA2}{D02B1F72-3407-48AE-BA88-E8213C6761F1}]"
    )

    gpt_ini = "\n".join(["[General]", "gPCMachineExtensionNames=", "Version=8", ""])
    expected = "\r\n".join(
        [
            "[General]",
            "gPCMachineExtensionNames=[{35378EAC-683F-11D2-A89A-00C04FBBCFA2}{D02B1F72-3407-48AE-BA88-E8213C6761F1}]",
            "Version=9",
            "",
        ]
    )

    with pytest.helpers.temp_file(
        "Registry.pol"
    ) as reg_pol_file, pytest.helpers.temp_file("gpt.ini") as gpt_ini_file:
        # We're using salt.utils.file.fopen here because the temp_file helper
        # doesn't preserve line endings when writing the test file
        with salt.utils.files.fopen(str(gpt_ini_file), "wb") as fp:
            fp.write(gpt_ini.encode("utf-8"))
        win_lgpo_reg.write_reg_pol_data(
            data_to_write=data_to_write,
            policy_file_path=str(reg_pol_file),
            gpt_extension=gpt_extension,
            gpt_extension_guid=gpt_extension_guid,
            gpt_ini_path=str(gpt_ini_file),
        )
        # We're using salt.utils.file.fopen here because the temp_file helper
        # doesn't preserve line endings when reading the test file
        with salt.utils.files.fopen(str(gpt_ini_file)) as fp:
            result = fp.read()

        assert result == expected


def test_issue_56769_mixed_line_endings():
    """
    Test that it handles a gpt.ini file with mixed line endings.
    Should create a gpt.ini with Windows-style line endings.
    """

    data_to_write = b"[\x00d\x00u\x00m\x00m\x00y\x00\\\x00d\x00a\x00t\x00a]\x00"
    gpt_extension = "gPCMachineExtensionNames"
    gpt_extension_guid = (
        "[{35378EAC-683F-11D2-A89A-00C04FBBCFA2}{D02B1F72-3407-48AE-BA88-E8213C6761F1}]"
    )

    gpt_ini = "[General]\ngPCMachineExtensionNames=\r\nVersion=8\n"
    expected = "\r\n".join(
        [
            "[General]",
            "gPCMachineExtensionNames=[{35378EAC-683F-11D2-A89A-00C04FBBCFA2}{D02B1F72-3407-48AE-BA88-E8213C6761F1}]",
            "Version=9",
            "",
        ]
    )

    with pytest.helpers.temp_file(
        "Registry.pol"
    ) as reg_pol_file, pytest.helpers.temp_file("gpt.ini") as gpt_ini_file:
        # We're using salt.utils.file.fopen here because the temp_file helper
        # doesn't preserve line endings when writing the test file
        with salt.utils.files.fopen(str(gpt_ini_file), "wb") as fp:
            fp.write(gpt_ini.encode("utf-8"))
        win_lgpo_reg.write_reg_pol_data(
            data_to_write=data_to_write,
            policy_file_path=str(reg_pol_file),
            gpt_extension=gpt_extension,
            gpt_extension_guid=gpt_extension_guid,
            gpt_ini_path=str(gpt_ini_file),
        )
        # We're using salt.utils.file.fopen here because the temp_file helper
        # doesn't preserve line endings when reading the test file
        with salt.utils.files.fopen(str(gpt_ini_file)) as fp:
            result = fp.read()

        assert result == expected



import functools
import logging
import random

import pytest
from saltfactories.utils import random_string

import salt.utils.files
import salt.utils.network
import salt.utils.path
from salt._compat import ipaddress
from salt.exceptions import CommandExecutionError

pytest.importorskip("docker")

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("docker", "dockerd", check_all=False),
]


IPV6_ENABLED = bool(salt.utils.network.ip_addrs6(include_loopback=True))


class Network:
    def __init__(self, name, **kwargs):
        self.kwargs = kwargs
        self.name = name
        try:
            self.net = ipaddress.ip_network(self.kwargs["subnet"])
            self._rand_indexes = random.sample(
                range(2, self.net.num_addresses - 1), self.net.num_addresses - 3
            )
            self.ip_arg = f"ipv{self.net.version}_address"
        except KeyError:
            # No explicit subnet passed
            self.net = self.ip_arg = None

    def __getitem__(self, index):
        try:
            return self.net[self._rand_indexes[index]].compressed
        except (TypeError, AttributeError):
            raise ValueError(
                "Indexing not supported for networks without a custom subnet"
            )

    @staticmethod
    def arg_map(arg_name):
        ret = {
            "ipv4_address": "IPv4Address",
            "ipv6_address": "IPv6Address",
            "links": "Links",
            "aliases": "Aliases",
        }
        return ret[arg_name]

    @property
    def subnet(self):
        try:
            return self.net.compressed
        except AttributeError:
            return None

    @property
    def gateway(self):
        try:
            return self.kwargs["gateway"]
        except KeyError:
            try:
                return self.net[1].compressed
            except (AttributeError, IndexError):
                return None


class CreateNetwork:
    """
    Generate a network for the test. Information about the network will be
    passed to the wrapped function.
    """

    def __init__(self, docker_module, **kwargs):
        self.docker_module = docker_module
        self.network = Network(random_string("salt_net_"), **kwargs)
        if self.network.net is not None:
            if "enable_ipv6" not in kwargs:
                kwargs["enable_ipv6"] = self.network.net.version == 6
        self.kwargs = kwargs

    def __enter__(self):
        return self.network

    def __exit__(self, *_):
        try:
            self.docker_module.disconnect_all_containers_from_network(self.network.name)
        except CommandExecutionError as exc:
            if "404" not in str(exc):
                raise exc from None
        else:
            self.docker_module.remove_network(self.network.name)
        self.network = None


@pytest.fixture(scope="module")
def container(salt_factories, state_tree):

    factory = salt_factories.get_container(
        random_string("docker-network-"),
        image_name="ghcr.io/saltstack/salt-ci-containers/busybox:musl",
        container_run_kwargs={
            "entrypoint": "tail -f /dev/null",
        },
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
    )

    with factory.started():
        yield factory


@pytest.fixture
def docker(modules):
    return modules.docker


@pytest.fixture
def docker_network(states):
    return states.docker_network


@pytest.fixture
def network(docker):
    return functools.partial(CreateNetwork, docker)


@pytest.fixture
def existing_network(network, docker_network):
    with network() as net:
        ret = docker_network.present(name=net.name)
        assert ret.result is True
        assert ret.changes
        assert ret.changes == {"created": True}
        try:
            yield net
        finally:
            docker_network.absent(name=net.name)


@pytest.fixture
def existing_network_with_container(network, docker_network, container):
    with network() as net:
        ret = docker_network.present(name=net.name, containers=[container.name])
        assert ret.result is True
        assert ret.changes
        assert ret.changes == {"created": True, "connected": [container.name]}
        try:
            yield net
        finally:
            docker_network.absent(name=net.name)


def test_absent(docker_network, existing_network):
    ret = docker_network.absent(name=existing_network.name)
    assert ret.result is True
    assert ret.changes
    assert ret.changes == {"removed": True}
    assert ret.comment
    assert ret.comment == f"Removed network '{existing_network.name}'"


def test_absent_with_disconnected_container(
    docker_network, container, existing_network_with_container
):
    ret = docker_network.absent(name=existing_network_with_container.name)
    assert ret.result is True
    assert ret.changes
    assert ret.changes == {"removed": True, "disconnected": [container.name]}
    assert ret.comment == "Removed network '{}'".format(
        existing_network_with_container.name
    )


def test_absent_when_not_present(network, docker_network):
    with network() as net:
        ret = docker_network.absent(name=net.name)
        assert ret.result is True
        assert not ret.changes
        assert ret.comment == f"Network '{net.name}' already absent"


def test_present(docker, network, docker_network):
    with network() as net:
        ret = docker_network.present(name=net.name)
        assert ret.result is True
        assert ret.changes
        assert ret.changes == {"created": True}
        assert ret.comment == f"Network '{net.name}' created"

        # Now check to see that the network actually exists. If it doesn't,
        # this next function call will raise an exception.
        docker.inspect_network(net.name)


def test_present_with_containers(network, docker, docker_network, container):
    with network() as net:
        ret = docker_network.present(name=net.name, containers=[container.name])
        assert ret.result is True
        assert ret.changes
        assert ret.changes == {"created": True, "connected": [container.name]}
        assert ret.comment == f"Network '{net.name}' created"

        # Now check to see that the network actually exists. If it doesn't,
        # this next function call will raise an exception.
        docker.inspect_network(net.name)


@pytest.mark.parametrize("reconnect", [True, False])
def test_present_with_reconnect(network, docker, docker_network, container, reconnect):
    """
    Test reconnecting with containers not passed to state
    """
    with network() as net:
        ret = docker_network.present(name=net.name, driver="bridge")
        assert ret.result is True
        assert ret.changes
        assert ret.changes == {"created": True}
        assert ret.comment == f"Network '{net.name}' created"

        # Connect the container
        docker.connect_container_to_network(container.name, net.name)

        # Change the driver to force the network to be replaced
        ret = docker_network.present(
            name=net.name, driver="macvlan", reconnect=reconnect
        )
        assert ret.result is True
        assert ret.changes
        assert ret.changes == {
            "recreated": True,
            "reconnected" if reconnect else "disconnected": [container.name],
            net.name: {
                "Driver": {
                    "old": "bridge",
                    "new": "macvlan",
                },
            },
        }
        assert ret.comment == "Network '{}' was replaced with updated config".format(
            net.name
        )


def test_present_internal(network, docker, docker_network):
    with network() as net:
        ret = docker_network.present(name=net.name, internal=True)
        assert ret.result is True
        net_info = docker.inspect_network(net.name)
        assert net_info["Internal"] is True


def test_present_labels(network, docker, docker_network):
    # Test a mix of different ways of specifying labels
    with network() as net:
        ret = docker_network.present(
            name=net.name,
            labels=["foo", "bar=baz", {"hello": "world"}],
        )
        assert ret.result is True
        net_info = docker.inspect_network(net.name)
        assert net_info["Labels"] == {"foo": "", "bar": "baz", "hello": "world"}


@pytest.mark.skipif(IPV6_ENABLED is False, reason="IPv6 not enabled")
def test_present_enable_ipv6(network, docker, docker_network):
    with network(subnet="10.247.197.96/27") as net1:
        with network(subnet="fe3f:2180:26:1::/123") as net2:
            ret = docker_network.present(
                name=net1.name,
                enable_ipv6=True,
                ipam_pools=[
                    {"subnet": net1.subnet},
                    {"subnet": net2.subnet},
                ],
            )
            assert ret.result is True
            net_info = docker.inspect_network(net1.name)
            assert net_info["EnableIPv6"] is True


def test_present_attachable(network, docker, docker_network, grains):
    if grains["os_family"] == "RedHat" and grains.get("osmajorrelease", 0) <= 7:
        pytest.skip("Cannot reliably manage attachable on RHEL <= 7")

    with network() as net:
        ret = docker_network.present(name=net.name, attachable=True)
        assert ret.result is True

        net_info = docker.inspect_network(net.name)
        assert net_info["Attachable"] is True


@pytest.mark.skipif(True, reason="Skip until we can set up docker swarm testing")
def test_present_scope(network, docker, docker_network):
    with network() as net:
        ret = docker_network.present(name=net.name, scope="global")
        assert ret.result is True

        net_info = docker.inspect_network(net.name)
        assert net_info["Scope"] == "global"


@pytest.mark.skipif(True, reason="Skip until we can set up docker swarm testing")
def test_present_ingress(network, docker, docker_network):
    with network() as net:
        ret = docker_network.present(name=net.name, ingress=True)
        assert ret.result is True

        net_info = docker.inspect_network(net.name)
        assert net_info["Ingress"] is True


def test_present_with_custom_ipv4(network, docker, docker_network):
    # First run will test passing the IPAM arguments individually
    with network(subnet="10.247.197.96/27") as net1, network(
        subnet="10.247.197.128/27"
    ) as net2:
        ret = docker_network.present(
            name=net1.name,
            subnet=net1.subnet,
            gateway=net1.gateway,
        )
        assert ret.result is True

        # Second run will pass them in the ipam_pools argument
        ret = docker_network.present(
            name=net1.name,  # We want to keep the same network name
            ipam_pools=[{"subnet": net2.subnet, "gateway": net2.gateway}],
        )
        assert ret.result is True

        # Docker requires there to be IPv4, even when only an IPv6 subnet was
        # provided. So, there will be both an IPv4 and IPv6 pool in the
        # configuration.
        expected = {
            "recreated": True,
            net1.name: {
                "IPAM": {
                    "Config": {
                        "old": [{"Subnet": net1.subnet, "Gateway": net1.gateway}],
                        "new": [{"Subnet": net2.subnet, "Gateway": net2.gateway}],
                    }
                }
            },
        }
        assert ret.changes
        assert ret.changes == expected
        assert ret.comment == "Network '{}' was replaced with updated config".format(
            net1.name
        )


@pytest.mark.skipif(IPV6_ENABLED is False, reason="IPv6 not enabled")
def test_present_with_custom_ipv6(network, docker, docker_network):
    with network(subnet="10.247.197.96/27") as ipv4_net, network(
        subnet="fe3f:2180:26:1::/123"
    ) as ipv6_net1, network(subnet="fe3f:2180:26:1::20/123") as ipv6_net2:
        ret = docker_network.present(
            name=ipv4_net.name,
            enable_ipv6=True,
            ipam_pools=[
                {"subnet": ipv4_net.subnet, "gateway": ipv4_net.gateway},
                {"subnet": ipv6_net1.subnet, "gateway": ipv6_net1.gateway},
            ],
        )
        assert ret.result is True

        ret = docker_network.present(
            name=ipv4_net.name,  # We want to keep the same network name
            enable_ipv6=True,
            ipam_pools=[
                {"subnet": ipv4_net.subnet, "gateway": ipv4_net.gateway},
                {"subnet": ipv6_net2.subnet, "gateway": ipv6_net2.gateway},
            ],
        )
        assert ret.result is True

        # Docker requires there to be IPv4, even when only an IPv6 subnet was
        # provided. So, there will be both an IPv4 and IPv6 pool in the
        # configuration.
        expected = {
            "recreated": True,
            ipv4_net.name: {
                "IPAM": {
                    "Config": {
                        "old": [
                            {"Subnet": ipv4_net.subnet, "Gateway": ipv4_net.gateway},
                            {"Subnet": ipv6_net1.subnet, "Gateway": ipv6_net1.gateway},
                        ],
                        "new": [
                            {"Subnet": ipv4_net.subnet, "Gateway": ipv4_net.gateway},
                            {"Subnet": ipv6_net2.subnet, "Gateway": ipv6_net2.gateway},
                        ],
                    }
                }
            },
        }
        assert ret.changes
        assert ret.changes == expected
        assert ret.comment == "Network '{}' was replaced with updated config".format(
            ipv4_net.name
        )


def test_bridge_dupname_update(network, docker, docker_network):
    # com.docker.network.bridge.name can not have names over 15 chars. so grab the last 8
    with network(subnet="10.247.197.96/27") as net:
        ret = docker_network.present(
            name=net.name,
            subnet=net.subnet,
            driver="bridge",
            driver_opts=[{"com.docker.network.bridge.name": net.name[-8:]}],
        )
        assert ret.result is True
        # Second run to make sure everything is still fine.
        ret = docker_network.present(
            name=net.name,
            subnet=net.subnet,
            driver="bridge",
            driver_opts=[{"com.docker.network.bridge.name": net.name[-8:]}],
        )
        assert ret.result is True
        assert not ret.changes
        assert (
            ret.comment
            == "Network '{}' already exists, and is configured as specified".format(
                net.name
            )
        )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.system as system
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="Skip system functions on Windows"),
]


@pytest.fixture
def configure_loader_modules():
    return {system: {}}


def test_halt():
    """
    Test to halt a running system
    """
    with patch.dict(system.__salt__, {"cmd.run": MagicMock(return_value="A")}):
        assert system.halt() == "A"


def test_init():
    """
    Test to change the system runlevel on sysV compatible systems
    """
    with patch.dict(system.__salt__, {"cmd.run": MagicMock(return_value="A")}):
        assert system.init("r") == "A"


def test_poweroff():
    """
    Test to poweroff a running system
    """
    with patch.dict(system.__salt__, {"cmd.run": MagicMock(return_value="A")}):
        assert system.poweroff() == "A"


def test_reboot():
    """
    Test to reboot the system with shutdown -r
    """
    cmd_mock = MagicMock(return_value="A")
    with patch.dict(system.__salt__, {"cmd.run": cmd_mock}):
        assert system.reboot() == "A"
    cmd_mock.assert_called_with(["shutdown", "-r", "now"], python_shell=False)


def test_reboot_with_delay():
    """
    Test to reboot the system using shutdown -r with a delay
    """
    cmd_mock = MagicMock(return_value="A")
    with patch.dict(system.__salt__, {"cmd.run": cmd_mock}):
        assert system.reboot(at_time=5) == "A"
    cmd_mock.assert_called_with(["shutdown", "-r", "5"], python_shell=False)


def test_shutdown():
    """
    Test to shutdown a running system
    """
    cmd_mock = MagicMock(return_value="A")
    with patch.dict(system.__salt__, {"cmd.run": cmd_mock}), patch(
        "salt.utils.platform.is_freebsd", MagicMock(return_value=False)
    ), patch("salt.utils.platform.is_netbsd", MagicMock(return_value=False)), patch(
        "salt.utils.platform.is_openbsd", MagicMock(return_value=False)
    ):
        assert system.shutdown() == "A"
    cmd_mock.assert_called_with(["shutdown", "-h", "now"], python_shell=False)


def test_shutdown_freebsd():
    """
    Test to shutdown a running FreeBSD system
    """
    cmd_mock = MagicMock(return_value="A")
    with patch.dict(system.__salt__, {"cmd.run": cmd_mock}), patch(
        "salt.utils.platform.is_freebsd", MagicMock(return_value=True)
    ), patch("salt.utils.platform.is_netbsd", MagicMock(return_value=False)), patch(
        "salt.utils.platform.is_openbsd", MagicMock(return_value=False)
    ):
        assert system.shutdown() == "A"
    cmd_mock.assert_called_with(["shutdown", "-p", "now"], python_shell=False)


def test_shutdown_netbsd():
    """
    Test to shutdown a running NetBSD system
    """
    cmd_mock = MagicMock(return_value="A")
    with patch.dict(system.__salt__, {"cmd.run": cmd_mock}), patch(
        "salt.utils.platform.is_freebsd", MagicMock(return_value=False)
    ), patch("salt.utils.platform.is_netbsd", MagicMock(return_value=True)), patch(
        "salt.utils.platform.is_openbsd", MagicMock(return_value=False)
    ):
        assert system.shutdown() == "A"
    cmd_mock.assert_called_with(["shutdown", "-p", "now"], python_shell=False)


def test_shutdown_openbsd():
    """
    Test to shutdown a running OpenBSD system
    """
    cmd_mock = MagicMock(return_value="A")
    with patch.dict(system.__salt__, {"cmd.run": cmd_mock}), patch(
        "salt.utils.platform.is_freebsd", MagicMock(return_value=False)
    ), patch("salt.utils.platform.is_netbsd", MagicMock(return_value=False)), patch(
        "salt.utils.platform.is_openbsd", MagicMock(return_value=True)
    ):
        assert system.shutdown() == "A"
    cmd_mock.assert_called_with(["shutdown", "-p", "now"], python_shell=False)



"""
Module for controlling Jenkins

:depends: python-jenkins

.. versionadded:: 2016.3.0

:depends: python-jenkins_ Python module (not to be confused with jenkins_)

.. _python-jenkins: https://pypi.python.org/pypi/python-jenkins
.. _jenkins: https://pypi.python.org/pypi/jenkins

:configuration: This module can be used by either passing an api key and version
    directly or by specifying both in a configuration profile in the salt
    master/minion config.

    For example:

    .. code-block:: yaml

        jenkins:
          api_key: peWcBiMOS9HrZG15peWcBiMOS9HrZG15
"""

import logging

import salt.utils.files
import salt.utils.stringutils

# pylint: disable=import-error,no-name-in-module,redefined-builtin
from salt.exceptions import CommandExecutionError, SaltInvocationError

try:
    import jenkins

    HAS_JENKINS = True
except ImportError:
    HAS_JENKINS = False


# pylint: enable=import-error,no-name-in-module

log = logging.getLogger(__name__)

__virtualname__ = "jenkins"


def __virtual__():
    """
    Return virtual name of the module.

    :return: The virtual name of the module.
    """
    if HAS_JENKINS:
        if hasattr(jenkins, "Jenkins"):
            return __virtualname__
        else:
            return (
                False,
                "The wrong Python module appears to be installed. Please "
                "make sure that 'python-jenkins' is installed, not "
                "'jenkins'.",
            )
    return (
        False,
        "The jenkins execution module cannot be loaded: "
        "python-jenkins is not installed.",
    )


def _connect():
    """
    Return server object used to interact with Jenkins.

    :return: server object used to interact with Jenkins
    """
    jenkins_url = (
        __salt__["config.get"]("jenkins.url")
        or __salt__["config.get"]("jenkins:url")
        or __salt__["pillar.get"]("jenkins.url")
    )

    jenkins_user = (
        __salt__["config.get"]("jenkins.user")
        or __salt__["config.get"]("jenkins:user")
        or __salt__["pillar.get"]("jenkins.user")
    )

    jenkins_password = (
        __salt__["config.get"]("jenkins.password")
        or __salt__["config.get"]("jenkins:password")
        or __salt__["pillar.get"]("jenkins.password")
    )

    if not jenkins_url:
        raise SaltInvocationError("No Jenkins URL found.")

    return jenkins.Jenkins(
        jenkins_url, username=jenkins_user, password=jenkins_password
    )


def _retrieve_config_xml(config_xml, saltenv):
    """
    Helper to cache the config XML and raise a CommandExecutionError if we fail
    to do so. If we successfully cache the file, return the cached path.
    """
    ret = __salt__["cp.cache_file"](config_xml, saltenv)

    if not ret:
        raise CommandExecutionError(f"Failed to retrieve {config_xml}")

    return ret


def run(script):
    """
    .. versionadded:: 2017.7.0

    Execute a script on the jenkins master

    :param script: The script

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.run 'Jenkins.instance.doSafeRestart()'

    """

    server = _connect()
    return server.run_script(script)


def get_version():
    """
    Return version of Jenkins

    :return: The version of Jenkins

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.get_version
    """

    server = _connect()

    version = server.get_version()
    if version:
        return version
    return False


def get_jobs():
    """
    Return the currently configured jobs.

    :return: The currently configured jobs.

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.get_jobs
    """

    server = _connect()

    jobs = server.get_jobs()
    if jobs:
        return jobs
    return {}


def job_exists(name=None):
    """
    Check whether the job exists in configured Jenkins jobs.

    :param name: The name of the job is check if it exists.
    :return: True if job exists, False if job does not exist.

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.job_exists jobname

    """
    if not name:
        raise SaltInvocationError("Required parameter 'name' is missing")

    server = _connect()
    if server.job_exists(name):
        return True
    else:
        return False


def get_job_info(name=None):
    """
    Return information about the Jenkins job.

    :param name: The name of the job is check if it exists.
    :return: Information about the Jenkins job.

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.get_job_info jobname

    """
    if not name:
        raise SaltInvocationError("Required parameter 'name' is missing")

    server = _connect()

    if not job_exists(name):
        raise CommandExecutionError(f"Job '{name}' does not exist")

    job_info = server.get_job_info(name)
    if job_info:
        return job_info
    return False


def build_job(name=None, parameters=None):
    """
    Initiate a build for the provided job.

    :param name: The name of the job is check if it exists.
    :param parameters: Parameters to send to the job.
    :return: True is successful, otherwise raise an exception.

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.build_job jobname

    """
    if not name:
        raise SaltInvocationError("Required parameter 'name' is missing")

    server = _connect()

    if not job_exists(name):
        raise CommandExecutionError(f"Job '{name}' does not exist.")

    try:
        server.build_job(name, parameters)
    except jenkins.JenkinsException as err:
        raise CommandExecutionError(f"Encountered error building job '{name}': {err}")
    return True


def create_job(name=None, config_xml=None, saltenv="base"):
    """
    Return the configuration file.

    :param name: The name of the job is check if it exists.
    :param config_xml: The configuration file to use to create the job.
    :param saltenv: The environment to look for the file in.
    :return: The configuration file used for the job.

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.create_job jobname

        salt '*' jenkins.create_job jobname config_xml='salt://jenkins/config.xml'

    """
    if not name:
        raise SaltInvocationError("Required parameter 'name' is missing")

    if job_exists(name):
        raise CommandExecutionError(f"Job '{name}' already exists")

    if not config_xml:
        config_xml = jenkins.EMPTY_CONFIG_XML
    else:
        config_xml_file = _retrieve_config_xml(config_xml, saltenv)

        with salt.utils.files.fopen(config_xml_file) as _fp:
            config_xml = salt.utils.stringutils.to_unicode(_fp.read())

    server = _connect()
    try:
        server.create_job(name, config_xml)
    except jenkins.JenkinsException as err:
        raise CommandExecutionError(f"Encountered error creating job '{name}': {err}")
    return config_xml


def update_job(name=None, config_xml=None, saltenv="base"):
    """
    Return the updated configuration file.

    :param name: The name of the job is check if it exists.
    :param config_xml: The configuration file to use to create the job.
    :param saltenv: The environment to look for the file in.
    :return: The configuration file used for the job.

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.update_job jobname

        salt '*' jenkins.update_job jobname config_xml='salt://jenkins/config.xml'

    """
    if not name:
        raise SaltInvocationError("Required parameter 'name' is missing")

    if not config_xml:
        config_xml = jenkins.EMPTY_CONFIG_XML
    else:
        config_xml_file = _retrieve_config_xml(config_xml, saltenv)

        with salt.utils.files.fopen(config_xml_file) as _fp:
            config_xml = salt.utils.stringutils.to_unicode(_fp.read())

    server = _connect()
    try:
        server.reconfig_job(name, config_xml)
    except jenkins.JenkinsException as err:
        raise CommandExecutionError(f"Encountered error updating job '{name}': {err}")
    return config_xml


def delete_job(name=None):
    """
    Return true is job is deleted successfully.

    :param name: The name of the job to delete.
    :return: Return true if job is deleted successfully.

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.delete_job jobname

    """
    if not name:
        raise SaltInvocationError("Required parameter 'name' is missing")

    server = _connect()

    if not job_exists(name):
        raise CommandExecutionError(f"Job '{name}' does not exist")

    try:
        server.delete_job(name)
    except jenkins.JenkinsException as err:
        raise CommandExecutionError(f"Encountered error deleting job '{name}': {err}")
    return True


def enable_job(name=None):
    """
    Return true is job is enabled successfully.

    :param name: The name of the job to enable.
    :return: Return true if job is enabled successfully.

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.enable_job jobname

    """
    if not name:
        raise SaltInvocationError("Required parameter 'name' is missing")

    server = _connect()

    if not job_exists(name):
        raise CommandExecutionError(f"Job '{name}' does not exist")

    try:
        server.enable_job(name)
    except jenkins.JenkinsException as err:
        raise CommandExecutionError(f"Encountered error enabling job '{name}': {err}")
    return True


def disable_job(name=None):
    """
    Return true is job is disabled successfully.

    :param name: The name of the job to disable.
    :return: Return true if job is disabled successfully.

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.disable_job jobname

    """

    if not name:
        raise SaltInvocationError("Required parameter 'name' is missing")

    server = _connect()

    if not job_exists(name):
        raise CommandExecutionError(f"Job '{name}' does not exist")

    try:
        server.disable_job(name)
    except jenkins.JenkinsException as err:
        raise CommandExecutionError(f"Encountered error disabling job '{name}': {err}")
    return True


def job_status(name=None):
    """
    Return the current status, enabled or disabled, of the job.

    :param name: The name of the job to return status for
    :return: Return true if enabled or false if disabled.

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.job_status jobname

    """

    if not name:
        raise SaltInvocationError("Required parameter 'name' is missing")

    server = _connect()

    if not job_exists(name):
        raise CommandExecutionError(f"Job '{name}' does not exist")

    return server.get_job_info("empty")["buildable"]


def get_job_config(name=None):
    """
    Return the current job configuration for the provided job.

    :param name: The name of the job to return the configuration for.
    :return: The configuration for the job specified.

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.get_job_config jobname

    """

    if not name:
        raise SaltInvocationError("Required parameter 'name' is missing")

    server = _connect()

    if not job_exists(name):
        raise CommandExecutionError(f"Job '{name}' does not exist")

    job_info = server.get_job_config(name)
    return job_info


def plugin_installed(name):
    """
    .. versionadded:: 2016.11.0

    Return if the plugin is installed for the provided plugin name.

    :param name: The name of the parameter to confirm installation.
    :return: True if plugin exists, False if plugin does not exist.

    CLI Example:

    .. code-block:: bash

        salt '*' jenkins.plugin_installed pluginName

    """

    server = _connect()
    plugins = server.get_plugins()

    exists = [plugin for plugin in plugins.keys() if name in plugin]

    if exists:
        return True
    else:
        return False



"""
The generic libcloud template used to create the connections and deploy the
cloud virtual machines
"""

import logging
import os

import salt.client
import salt.config
import salt.utils.cloud
import salt.utils.data
import salt.utils.event
from salt.exceptions import SaltCloudNotFound, SaltCloudSystemExit

# pylint: disable=W0611
try:
    import libcloud
    from libcloud.compute.deployment import MultiStepDeployment, ScriptDeployment
    from libcloud.compute.providers import get_driver
    from libcloud.compute.types import NodeState, Provider

    HAS_LIBCLOUD = True
    LIBCLOUD_VERSION_INFO = tuple(
        int(part)
        for part in libcloud.__version__.replace("-", ".")
        .replace("rc", ".")
        .split(".")[:3]
    )

except ImportError:
    HAS_LIBCLOUD = False
    LIBCLOUD_VERSION_INFO = (1000,)
# pylint: enable=W0611


log = logging.getLogger(__name__)


LIBCLOUD_MINIMAL_VERSION = (1, 5, 0)


def check_libcloud_version(reqver=LIBCLOUD_MINIMAL_VERSION, why=None):
    """
    Compare different libcloud versions
    """
    if not HAS_LIBCLOUD:
        return False

    if not isinstance(reqver, (list, tuple)):
        raise RuntimeError(
            "'reqver' needs to passed as a tuple or list, i.e., (1, 5, 0)"
        )

    if LIBCLOUD_VERSION_INFO >= reqver:
        return libcloud.__version__

    errormsg = (
        "Your version of libcloud is {}. salt-cloud requires >= libcloud {}".format(
            libcloud.__version__, ".".join([str(num) for num in reqver])
        )
    )
    if why:
        errormsg += f" for {why}"
    errormsg += ". Please upgrade."
    raise ImportError(errormsg)


def get_node(conn, name):
    """
    Return a libcloud node for the named VM
    """
    nodes = conn.list_nodes()
    for node in nodes:
        if node.name == name:
            __utils__["cloud.cache_node"](
                salt.utils.data.simple_types_filter(node.__dict__),
                __active_provider_name__,
                __opts__,
            )
            return node


def avail_locations(conn=None, call=None):
    """
    Return a dict of all available VM locations on the cloud provider with
    relevant data
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_locations function must be called with "
            "-f or --function, or with the --list-locations option"
        )

    if not conn:
        conn = get_conn()  # pylint: disable=E0602

    locations = conn.list_locations()
    ret = {}
    for img in locations:
        img_name = str(img.name)

        ret[img_name] = {}
        for attr in dir(img):
            if attr.startswith("_") or attr == "driver":
                continue

            attr_value = getattr(img, attr)
            ret[img_name][attr] = attr_value

    return ret


def avail_images(conn=None, call=None):
    """
    Return a dict of all available VM images on the cloud provider with
    relevant data
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_images function must be called with "
            "-f or --function, or with the --list-images option"
        )

    if not conn:
        conn = get_conn()  # pylint: disable=E0602

    images = conn.list_images()
    ret = {}
    for img in images:
        img_name = str(img.name)

        ret[img_name] = {}
        for attr in dir(img):
            if attr.startswith("_") or attr in ("driver", "get_uuid"):
                continue
            attr_value = getattr(img, attr)
            ret[img_name][attr] = attr_value
    return ret


def avail_sizes(conn=None, call=None):
    """
    Return a dict of all available VM images on the cloud provider with
    relevant data
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_sizes function must be called with "
            "-f or --function, or with the --list-sizes option"
        )

    if not conn:
        conn = get_conn()  # pylint: disable=E0602

    sizes = conn.list_sizes()
    ret = {}
    for size in sizes:
        size_name = str(size.name)

        ret[size_name] = {}
        for attr in dir(size):
            if attr.startswith("_") or attr in ("driver", "get_uuid"):
                continue

            try:
                attr_value = getattr(size, attr)
            except Exception:  # pylint: disable=broad-except
                pass

            ret[size_name][attr] = attr_value
    return ret


def get_location(conn, vm_):
    """
    Return the location object to use
    """
    locations = conn.list_locations()
    vm_location = salt.config.get_cloud_config_value("location", vm_, __opts__)
    for img in locations:
        img_id = str(img.id)
        img_name = str(img.name)

        if vm_location and vm_location in (img_id, img_name):
            return img

    raise SaltCloudNotFound(
        f"The specified location, '{vm_location}', could not be found."
    )


def get_image(conn, vm_):
    """
    Return the image object to use
    """
    images = conn.list_images()
    vm_image = salt.config.get_cloud_config_value("image", vm_, __opts__)

    for img in images:
        img_id = str(img.id)
        img_name = str(img.name)

        if vm_image and vm_image in (img_id, img_name):
            return img

    raise SaltCloudNotFound(f"The specified image, '{vm_image}', could not be found.")


def get_size(conn, vm_):
    """
    Return the VM's size object
    """
    sizes = conn.list_sizes()
    vm_size = salt.config.get_cloud_config_value("size", vm_, __opts__)
    if not vm_size:
        return sizes[0]

    for size in sizes:
        if vm_size and str(vm_size) in (
            str(size.id),
            str(size.name),
        ):
            return size
    raise SaltCloudNotFound(f"The specified size, '{vm_size}', could not be found.")


def script(vm_):
    """
    Return the script deployment object
    """
    return ScriptDeployment(
        salt.utils.cloud.os_script(
            salt.config.get_cloud_config_value("os", vm_, __opts__),
            vm_,
            __opts__,
            salt.utils.cloud.salt_config_to_yaml(
                salt.utils.cloud.minion_config(__opts__, vm_)
            ),
        )
    )


def destroy(name, conn=None, call=None):
    """
    Delete a single VM
    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )

    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        f"salt/cloud/{name}/destroying",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    if not conn:
        conn = get_conn()  # pylint: disable=E0602

    node = get_node(conn, name)
    profiles = get_configured_provider()["profiles"]  # pylint: disable=E0602
    if node is None:
        log.error("Unable to find the VM %s", name)
    profile = None
    if "metadata" in node.extra and "profile" in node.extra["metadata"]:
        profile = node.extra["metadata"]["profile"]

    flush_mine_on_destroy = False
    if profile and profile in profiles and "flush_mine_on_destroy" in profiles[profile]:
        flush_mine_on_destroy = profiles[profile]["flush_mine_on_destroy"]

    if flush_mine_on_destroy:
        log.info("Clearing Salt Mine: %s", name)

        mopts_ = salt.config.DEFAULT_MINION_OPTS
        conf_path = "/".join(__opts__["conf_file"].split("/")[:-1])
        mopts_.update(salt.config.minion_config(os.path.join(conf_path, "minion")))
        with salt.client.get_local_client(mopts=mopts_) as client:
            minions = client.cmd(name, "mine.flush")

    log.info("Clearing Salt Mine: %s, %s", name, flush_mine_on_destroy)
    log.info("Destroying VM: %s", name)
    ret = conn.destroy_node(node)
    if ret:
        log.info("Destroyed VM: %s", name)
        # Fire destroy action
        __utils__["cloud.fire_event"](
            "event",
            "destroyed instance",
            f"salt/cloud/{name}/destroyed",
            args={"name": name},
            sock_dir=__opts__["sock_dir"],
            transport=__opts__["transport"],
        )
        if __opts__["delete_sshkeys"] is True:
            public_ips = getattr(node, __opts__.get("ssh_interface", "public_ips"))
            if public_ips:
                salt.utils.cloud.remove_sshkey(public_ips[0])

            private_ips = getattr(node, __opts__.get("ssh_interface", "private_ips"))
            if private_ips:
                salt.utils.cloud.remove_sshkey(private_ips[0])

        if __opts__.get("update_cachedir", False) is True:
            __utils__["cloud.delete_minion_cachedir"](
                name, __active_provider_name__.split(":")[0], __opts__
            )

        return True

    log.error("Failed to Destroy VM: %s", name)
    return False


def reboot(name, conn=None):
    """
    Reboot a single VM
    """
    if not conn:
        conn = get_conn()  # pylint: disable=E0602

    node = get_node(conn, name)
    if node is None:
        log.error("Unable to find the VM %s", name)
    log.info("Rebooting VM: %s", name)
    ret = conn.reboot_node(node)
    if ret:
        log.info("Rebooted VM: %s", name)
        # Fire reboot action
        __utils__["cloud.fire_event"](
            "event",
            f"{name} has been rebooted",
            f"salt/cloud/{name}/rebooting",
            args={"name": name},
            sock_dir=__opts__["sock_dir"],
            transport=__opts__["transport"],
        )
        return True

    log.error("Failed to reboot VM: %s", name)
    return False


def list_nodes(conn=None, call=None):
    """
    Return a list of the VMs that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes function must be called with -f or --function."
        )

    if not conn:
        conn = get_conn()  # pylint: disable=E0602

    nodes = conn.list_nodes()
    ret = {}
    for node in nodes:
        ret[node.name] = {
            "id": node.id,
            "image": node.image,
            "name": node.name,
            "private_ips": node.private_ips,
            "public_ips": node.public_ips,
            "size": node.size,
            "state": str(node.state).upper(),
        }
    return ret


def list_nodes_full(conn=None, call=None):
    """
    Return a list of the VMs that are on the provider, with all fields
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_full function must be called with -f or --function."
        )

    if not conn:
        conn = get_conn()  # pylint: disable=E0602

    nodes = conn.list_nodes()
    ret = {}
    for node in nodes:
        pairs = {}
        for key, value in zip(node.__dict__, node.__dict__.values()):
            pairs[key] = value
        ret[node.name] = pairs
        del ret[node.name]["driver"]

    __utils__["cloud.cache_node_list"](
        ret, __active_provider_name__.split(":")[0], __opts__
    )
    return ret


def list_nodes_select(conn=None, call=None):
    """
    Return a list of the VMs that are on the provider, with select fields
    """
    if not conn:
        conn = get_conn()  # pylint: disable=E0602

    return salt.utils.cloud.list_nodes_select(
        list_nodes_full(conn, "function"),
        __opts__["query.selection"],
        call,
    )


def show_instance(name, call=None):
    """
    Show the details from the provider concerning an instance
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_instance action must be called with -a or --action."
        )

    nodes = list_nodes_full()
    __utils__["cloud.cache_node"](nodes[name], __active_provider_name__, __opts__)
    return nodes[name]


def conn_has_method(conn, method_name):
    """
    Find if the provided connection object has a specific method
    """
    if method_name in dir(conn):
        return True

    log.error("Method '%s' not yet supported!", method_name)
    return False



"""
Module to provide redis functionality to Salt

.. versionadded:: 2014.7.0

:configuration: This module requires the redis python module and uses the
    following defaults which may be overridden in the minion configuration:

.. code-block:: yaml

    redis.host: 'salt'
    redis.port: 6379
    redis.db: 0
    redis.password: None
"""

import salt.utils.args

try:
    import redis

    HAS_REDIS = True
except ImportError:
    HAS_REDIS = False

__virtualname__ = "redis"


def __virtual__():
    """
    Only load this module if redis python module is installed
    """
    if HAS_REDIS:
        return __virtualname__
    else:
        return (
            False,
            "The redis execution module failed to load: the redis python library is not"
            " available.",
        )


def _connect(host=None, port=None, db=None, password=None):
    """
    Returns an instance of the redis client
    """
    if not host:
        host = __salt__["config.option"]("redis.host")
    if not port:
        port = __salt__["config.option"]("redis.port")
    if not db:
        db = __salt__["config.option"]("redis.db")
    if not password:
        password = __salt__["config.option"]("redis.password")

    return redis.StrictRedis(host, port, db, password, decode_responses=True)


def _sconnect(host=None, port=None, password=None):
    """
    Returns an instance of the redis client
    """
    if host is None:
        host = __salt__["config.option"]("redis_sentinel.host", "localhost")
    if port is None:
        port = __salt__["config.option"]("redis_sentinel.port", 26379)
    if password is None:
        password = __salt__["config.option"]("redis_sentinel.password")

    return redis.StrictRedis(host, port, password=password, decode_responses=True)


def bgrewriteaof(host=None, port=None, db=None, password=None):
    """
    Asynchronously rewrite the append-only file

    CLI Example:

    .. code-block:: bash

        salt '*' redis.bgrewriteaof
    """
    server = _connect(host, port, db, password)
    return server.bgrewriteaof()


def bgsave(host=None, port=None, db=None, password=None):
    """
    Asynchronously save the dataset to disk

    CLI Example:

    .. code-block:: bash

        salt '*' redis.bgsave
    """
    server = _connect(host, port, db, password)
    return server.bgsave()


def config_get(pattern="*", host=None, port=None, db=None, password=None):
    """
    Get redis server configuration values

    CLI Example:

    .. code-block:: bash

        salt '*' redis.config_get
        salt '*' redis.config_get port
    """
    server = _connect(host, port, db, password)
    return server.config_get(pattern)


def config_set(name, value, host=None, port=None, db=None, password=None):
    """
    Set redis server configuration values

    CLI Example:

    .. code-block:: bash

        salt '*' redis.config_set masterauth luv_kittens
    """
    server = _connect(host, port, db, password)
    return server.config_set(name, value)


def dbsize(host=None, port=None, db=None, password=None):
    """
    Return the number of keys in the selected database

    CLI Example:

    .. code-block:: bash

        salt '*' redis.dbsize
    """
    server = _connect(host, port, db, password)
    return server.dbsize()


def delete(*keys, **connection_args):
    """
    Deletes the keys from redis, returns number of keys deleted

    CLI Example:

    .. code-block:: bash

        salt '*' redis.delete foo
    """
    # Get connection args from keywords if set
    conn_args = {}
    for arg in ["host", "port", "db", "password"]:
        if arg in connection_args:
            conn_args[arg] = connection_args[arg]

    server = _connect(**conn_args)
    return server.delete(*keys)


def exists(key, host=None, port=None, db=None, password=None):
    """
    Return true if the key exists in redis

    CLI Example:

    .. code-block:: bash

        salt '*' redis.exists foo
    """
    server = _connect(host, port, db, password)
    return server.exists(key)


def expire(key, seconds, host=None, port=None, db=None, password=None):
    """
    Set a keys time to live in seconds

    CLI Example:

    .. code-block:: bash

        salt '*' redis.expire foo 300
    """
    server = _connect(host, port, db, password)
    return server.expire(key, seconds)


def expireat(key, timestamp, host=None, port=None, db=None, password=None):
    """
    Set a keys expire at given UNIX time

    CLI Example:

    .. code-block:: bash

        salt '*' redis.expireat foo 1400000000
    """
    server = _connect(host, port, db, password)
    return server.expireat(key, timestamp)


def flushall(host=None, port=None, db=None, password=None):
    """
    Remove all keys from all databases

    CLI Example:

    .. code-block:: bash

        salt '*' redis.flushall
    """
    server = _connect(host, port, db, password)
    return server.flushall()


def flushdb(host=None, port=None, db=None, password=None):
    """
    Remove all keys from the selected database

    CLI Example:

    .. code-block:: bash

        salt '*' redis.flushdb
    """
    server = _connect(host, port, db, password)
    return server.flushdb()


def get_key(key, host=None, port=None, db=None, password=None):
    """
    Get redis key value

    CLI Example:

    .. code-block:: bash

        salt '*' redis.get_key foo
    """
    server = _connect(host, port, db, password)
    return server.get(key)


def hdel(key, *fields, **options):
    """
    Delete one of more hash fields.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hdel foo_hash bar_field1 bar_field2
    """
    host = options.get("host", None)
    port = options.get("port", None)
    database = options.get("db", None)
    password = options.get("password", None)
    server = _connect(host, port, database, password)
    return server.hdel(key, *fields)


def hexists(key, field, host=None, port=None, db=None, password=None):
    """
    Determine if a hash fields exists.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hexists foo_hash bar_field
    """
    server = _connect(host, port, db, password)
    return server.hexists(key, field)


def hget(key, field, host=None, port=None, db=None, password=None):
    """
    Get specific field value from a redis hash, returns dict

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hget foo_hash bar_field
    """
    server = _connect(host, port, db, password)
    return server.hget(key, field)


def hgetall(key, host=None, port=None, db=None, password=None):
    """
    Get all fields and values from a redis hash, returns dict

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hgetall foo_hash
    """
    server = _connect(host, port, db, password)
    return server.hgetall(key)


def hincrby(key, field, increment=1, host=None, port=None, db=None, password=None):
    """
    Increment the integer value of a hash field by the given number.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hincrby foo_hash bar_field 5
    """
    server = _connect(host, port, db, password)
    return server.hincrby(key, field, amount=increment)


def hincrbyfloat(
    key, field, increment=1.0, host=None, port=None, db=None, password=None
):
    """
    Increment the float value of a hash field by the given number.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hincrbyfloat foo_hash bar_field 5.17
    """
    server = _connect(host, port, db, password)
    return server.hincrbyfloat(key, field, amount=increment)


def hlen(key, host=None, port=None, db=None, password=None):
    """
    Returns number of fields of a hash.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hlen foo_hash
    """
    server = _connect(host, port, db, password)
    return server.hlen(key)


def hmget(key, *fields, **options):
    """
    Returns the values of all the given hash fields.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hmget foo_hash bar_field1 bar_field2
    """
    host = options.get("host", None)
    port = options.get("port", None)
    database = options.get("db", None)
    password = options.get("password", None)
    server = _connect(host, port, database, password)
    return server.hmget(key, *fields)


def hmset(key, **fieldsvals):
    """
    Sets multiple hash fields to multiple values.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hmset foo_hash bar_field1=bar_value1 bar_field2=bar_value2
    """
    host = fieldsvals.pop("host", None)
    port = fieldsvals.pop("port", None)
    database = fieldsvals.pop("db", None)
    password = fieldsvals.pop("password", None)
    server = _connect(host, port, database, password)
    return server.hmset(key, salt.utils.args.clean_kwargs(**fieldsvals))


def hset(key, field, value, host=None, port=None, db=None, password=None):
    """
    Set the value of a hash field.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hset foo_hash bar_field bar_value
    """
    server = _connect(host, port, db, password)
    return server.hset(key, field, value)


def hsetnx(key, field, value, host=None, port=None, db=None, password=None):
    """
    Set the value of a hash field only if the field does not exist.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hsetnx foo_hash bar_field bar_value
    """
    server = _connect(host, port, db, password)
    return server.hsetnx(key, field, value)


def hvals(key, host=None, port=None, db=None, password=None):
    """
    Return all the values in a hash.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hvals foo_hash bar_field1 bar_value1
    """
    server = _connect(host, port, db, password)
    return server.hvals(key)


def hscan(
    key, cursor=0, match=None, count=None, host=None, port=None, db=None, password=None
):
    """
    Incrementally iterate hash fields and associated values.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.hscan foo_hash match='field_prefix_*' count=1
    """
    server = _connect(host, port, db, password)
    return server.hscan(key, cursor=cursor, match=match, count=count)


def info(host=None, port=None, db=None, password=None):
    """
    Get information and statistics about the server

    CLI Example:

    .. code-block:: bash

        salt '*' redis.info
    """
    server = _connect(host, port, db, password)
    return server.info()


def keys(pattern="*", host=None, port=None, db=None, password=None):
    """
    Get redis keys, supports glob style patterns

    CLI Example:

    .. code-block:: bash

        salt '*' redis.keys
        salt '*' redis.keys test*
    """
    server = _connect(host, port, db, password)
    return server.keys(pattern)


def key_type(key, host=None, port=None, db=None, password=None):
    """
    Get redis key type

    CLI Example:

    .. code-block:: bash

        salt '*' redis.type foo
    """
    server = _connect(host, port, db, password)
    return server.type(key)


def lastsave(host=None, port=None, db=None, password=None):
    """
    Get the UNIX time in seconds of the last successful save to disk

    CLI Example:

    .. code-block:: bash

        salt '*' redis.lastsave
    """
    # Use of %s to get the timestamp is not supported by Python. The reason it
    # works is because it's passed to the system strftime which may not support
    # it. See: https://stackoverflow.com/a/11743262
    server = _connect(host, port, db, password)
    return int(server.lastsave().timestamp())


def llen(key, host=None, port=None, db=None, password=None):
    """
    Get the length of a list in Redis

    CLI Example:

    .. code-block:: bash

        salt '*' redis.llen foo_list
    """
    server = _connect(host, port, db, password)
    return server.llen(key)


def lrange(key, start, stop, host=None, port=None, db=None, password=None):
    """
    Get a range of values from a list in Redis

    CLI Example:

    .. code-block:: bash

        salt '*' redis.lrange foo_list 0 10
    """
    server = _connect(host, port, db, password)
    return server.lrange(key, start, stop)


def ping(host=None, port=None, db=None, password=None):
    """
    Ping the server, returns False on connection errors

    CLI Example:

    .. code-block:: bash

        salt '*' redis.ping
    """
    server = _connect(host, port, db, password)
    try:
        return server.ping()
    except redis.ConnectionError:
        return False


def save(host=None, port=None, db=None, password=None):
    """
    Synchronously save the dataset to disk

    CLI Example:

    .. code-block:: bash

        salt '*' redis.save
    """
    server = _connect(host, port, db, password)
    return server.save()


def set_key(key, value, host=None, port=None, db=None, password=None):
    """
    Set redis key value

    CLI Example:

    .. code-block:: bash

        salt '*' redis.set_key foo bar
    """
    server = _connect(host, port, db, password)
    return server.set(key, value)


def shutdown(host=None, port=None, db=None, password=None):
    """
    Synchronously save the dataset to disk and then shut down the server

    CLI Example:

    .. code-block:: bash

        salt '*' redis.shutdown
    """
    server = _connect(host, port, db, password)
    try:
        # Return false if unable to ping server
        server.ping()
    except redis.ConnectionError:
        return False

    server.shutdown()
    try:
        # This should fail now if the server is shutdown, which we want
        server.ping()
    except redis.ConnectionError:
        return True
    return False


def slaveof(
    master_host=None, master_port=None, host=None, port=None, db=None, password=None
):
    """
    Make the server a slave of another instance, or promote it as master

    CLI Example:

    .. code-block:: bash

        # Become slave of redis-n01.example.com:6379
        salt '*' redis.slaveof redis-n01.example.com 6379
        salt '*' redis.slaveof redis-n01.example.com
        # Become master
        salt '*' redis.slaveof
    """
    if master_host and not master_port:
        master_port = 6379
    server = _connect(host, port, db, password)
    return server.slaveof(master_host, master_port)


def smembers(key, host=None, port=None, db=None, password=None):
    """
    Get members in a Redis set

    CLI Example:

    .. code-block:: bash

        salt '*' redis.smembers foo_set
    """
    server = _connect(host, port, db, password)
    return list(server.smembers(key))


def time(host=None, port=None, db=None, password=None):
    """
    Return the current server UNIX time in seconds

    CLI Example:

    .. code-block:: bash

        salt '*' redis.time
    """
    server = _connect(host, port, db, password)
    return server.time()[0]


def zcard(key, host=None, port=None, db=None, password=None):
    """
    Get the length of a sorted set in Redis

    CLI Example:

    .. code-block:: bash

        salt '*' redis.zcard foo_sorted
    """
    server = _connect(host, port, db, password)
    return server.zcard(key)


def zrange(key, start, stop, host=None, port=None, db=None, password=None):
    """
    Get a range of values from a sorted set in Redis by index

    CLI Example:

    .. code-block:: bash

        salt '*' redis.zrange foo_sorted 0 10
    """
    server = _connect(host, port, db, password)
    return server.zrange(key, start, stop)


def sentinel_get_master_ip(master, host=None, port=None, password=None):
    """
    Get ip for sentinel master

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.sentinel_get_master_ip 'mymaster'
    """
    server = _sconnect(host, port, password)
    ret = server.sentinel_get_master_addr_by_name(master)
    return dict(list(zip(("master_host", "master_port"), ret)))


def get_master_ip(host=None, port=None, password=None):
    """
    Get host information about slave

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' redis.get_master_ip
    """
    server = _connect(host, port, password)
    srv_info = server.info()
    ret = (srv_info.get("master_host", ""), srv_info.get("master_port", ""))
    return dict(list(zip(("master_host", "master_port"), ret)))



"""
The ssh client wrapper system contains the routines that are used to alter
how executions are run in the salt-ssh system, this allows for state routines
to be easily rewritten to execute in a way that makes them do the same tasks
as ZeroMQ salt, but via ssh.
"""

import logging
from collections.abc import MutableMapping

import salt.client.ssh
import salt.loader
import salt.utils.data
import salt.utils.json
from salt.defaults import NOT_SET
from salt.exceptions import CommandExecutionError, SaltException

log = logging.getLogger(__name__)


class SSHException(SaltException):
    """
    Indicates general command failure via salt-ssh.
    """

    _error = ""

    def __init__(
        self, stdout, stderr, retcode, result=NOT_SET, parsed=None, *args, **kwargs
    ):
        super().__init__(stderr, *args, **kwargs)
        self.stdout = stdout
        self.stderr = self._filter_stderr(stderr)
        self.result = result
        self.parsed = parsed
        self.retcode = retcode
        if args:
            self._error = args.pop(0)
        super().__init__(self._error)

    def _filter_stderr(self, stderr):
        stderr_lines = []
        skip_next = False
        for line in stderr.splitlines():
            if skip_next:
                skip_next = False
                continue
            # Filter out deprecation warnings from stderr to the best of
            # our ability since they are irrelevant to the command output and cause noise.
            parts = line.split(":")
            if len(parts) > 2 and "DeprecationWarning" in parts[2]:
                # DeprecationWarnings print two lines, the second one being the
                # line that caused the warning.
                skip_next = True
                continue
            stderr_lines.append(line)
        return "\n".join(stderr_lines)

    def to_ret(self):
        ret = {
            "stdout": self.stdout,
            "stderr": self.stderr,
            "retcode": self.retcode,
            "parsed": self.parsed,
        }
        if self._error:
            ret["_error"] = self._error
        if self.result is not NOT_SET:
            ret["return"] = self.result
        return ret


class SSHCommandExecutionError(SSHException, CommandExecutionError):
    """
    Thrown whenever a non-zero exit code is returned.
    This was introduced to make the salt-ssh FunctionWrapper behave
    more like the usual one, in particular to force template rendering
    to stop when a function call results in an exception.
    """

    _error = "The command resulted in a non-zero exit code"

    def to_ret(self):
        if self.parsed and "local" in self.parsed:
            # Wrapped commands that indicate a non-zero retcode
            return self.parsed["local"]
        return super().to_ret()

    def __str__(self):
        if self.retcode > 0:
            return f"{self._error}: {self.stderr or self.stdout}"
        return self._error


class SSHPermissionDeniedError(SSHException):
    """
    Thrown when "Permission denied" is found in stderr
    """

    _error = "Permission denied"


class SSHReturnDecodeError(SSHException):
    """
    Thrown when JSON-decoding stdout fails and the retcode is 0 otherwise
    """

    _error = "Failed to return clean data"


class SSHMalformedReturnError(SSHException):
    """
    Thrown when a decoded return dict is not formed as
    {"local": {"return": ...}}
    """

    _error = "Return dict was malformed"


class LoadedMod:
    """
    This class is used as a proxy to a loaded wrapper module
    or the module part of a call to the target when
    a non-recommended syntax is used for loader access
    (like ``salt.grains.get`` or ``salt["grains"].get``).
    """

    __slots__ = ("mod", "wrapper")

    def __init__(self, mod, wrapper):
        self.mod = mod
        self.wrapper = wrapper

    def __getattr__(self, name):
        """
        Return the requested function.
        """
        try:
            return self.wrapper[f"{self.mod}.{name}"]
        except KeyError:
            # This  shouldn't happen since we wrap unknown calls to the target
            raise AttributeError(
                f"No attribute by the name of {name} was found on {self.mod}"
            )

    def __setitem__(self, name, value):
        """
        Set aliases for functions
        """
        self.wrapper[f"{self.mod}.{name}"] = value

    def __delitem__(self, name):
        """
        Remove aliases for functions
        """
        del self.wrapper[f"{self.mod}.{name}"]

    def __repr__(self):
        try:
            # Determine if we're representing a wrapper module or
            # an unknown execution module on the target.
            # We need to use the attribute since __getitem__ does not
            # allow module-level access.
            getattr(
                self.wrapper.wfuncs, self.mod
            )  # pylint: disable=pointless-statement
            prefix = self.wrapper.wfuncs.loaded_base_name + "."
            name = self.__class__.__name__
        except AttributeError:
            prefix = ""
            name = "SSHTargetMod"
        return f"<{name} module='{prefix}{self.mod}'>"


class FunctionWrapper(MutableMapping):
    """
    Create an object that acts like the salt function dict and makes function
    calls remotely via the SSH shell system
    """

    def __init__(
        self,
        opts,
        id_,
        host,
        wfuncs=None,
        mods=None,
        fsclient=None,
        aliases=None,
        minion_opts=None,
        **kwargs,
    ):
        super().__init__()
        self.wfuncs = wfuncs if wfuncs is not None else {}
        self.opts = opts
        self.mods = mods if isinstance(mods, dict) else {}
        self.kwargs = {"id_": id_, "host": host}
        self.fsclient = fsclient
        self.kwargs.update(kwargs)
        self.aliases = aliases
        if self.aliases is None:
            self.aliases = {}
        self.minion_opts = minion_opts

    def __contains__(self, key):
        """
        We need to implement a __contains__ method, othwerwise when someone
        does a contains comparison python assumes this is a sequence, and does
        __getitem__ keys 0 and up until IndexError
        """
        try:
            self[key]  # pylint: disable=pointless-statement
            return True
        except KeyError:
            return False

    def __getitem__(self, cmd):
        """
        Return the function call to simulate the salt local lookup system
        """
        if "." not in cmd:
            # Form of salt.cmd.run in Jinja -- it's expecting a subdictionary
            # containing only 'cmd' module calls
            # We don't know which modules are available on the target, so just
            # return the module namespace without any checks.
            return LoadedMod(cmd, self)

        if cmd in self.wfuncs:
            return self.wfuncs[cmd]

        if cmd in self.aliases:
            return self.aliases[cmd]

        def caller(*args, **kwargs):
            """
            The remote execution function
            """
            argv = [cmd]
            argv.extend([salt.utils.json.dumps(arg) for arg in args])
            argv.extend(
                [
                    "{}={}".format(
                        salt.utils.stringutils.to_str(key), salt.utils.json.dumps(val)
                    )
                    for key, val in kwargs.items()
                ]
            )
            single = salt.client.ssh.Single(
                self.opts,
                argv,
                mods=self.mods,
                disable_wipe=True,
                fsclient=self.fsclient,
                minion_opts=self.minion_opts,
                **self.kwargs,
            )
            stdout, stderr, retcode = single.cmd_block()
            return parse_ret(stdout, stderr, retcode, result_only=True)

        return caller

    def __setitem__(self, cmd, value):
        """
        Set aliases for functions
        """
        if "." not in cmd:
            # Form of salt.cmd.run in Jinja -- it's expecting a subdictionary
            # containing only 'cmd' module calls, in that case. We don't
            # support assigning directly to prefixes in this way
            raise KeyError(f"Cannot assign to module key {cmd} in the FunctionWrapper")

        if cmd in self.wfuncs:
            self.wfuncs[cmd] = value

        # Here was assume `value` is a `caller` function from __getitem__.
        # We save it as an alias and then can return it when referenced
        # later in __getitem__
        self.aliases[cmd] = value

    def __delitem__(self, cmd):
        """
        Remove aliases for functions
        """
        if "." not in cmd:
            # Form of salt.cmd.run in Jinja
            raise KeyError(f"Cannot delete module key {cmd} in the FunctionWrapper")

        if cmd in self.wfuncs:
            del self.wfuncs[cmd]

        del self.aliases[cmd]

    def __len__(self):
        """
        Return the count of wrapper modules and aliases.
        We don't know which modules will be available on the target.
        """
        return len(self.wfuncs) + len(self.aliases)

    def __iter__(self):
        """
        Iterate through wrapper modules and aliases.
        We don't know which modules will be available on the target.
        """
        yield from self.wfuncs
        yield from self.aliases

    def __getattr__(self, mod_or_func):
        """
        Ensure the behavior is similar to the usual LazyLoader regarding
        attribute access.
        """
        if mod_or_func.startswith("__") and mod_or_func.endswith("__"):
            # Don't pretend dunders are set.
            raise AttributeError(mod_or_func)
        try:
            return self.__getitem__(mod_or_func)
        except KeyError:
            raise AttributeError(mod_or_func)


def parse_ret(stdout, stderr, retcode, result_only=False):
    """
    Parse the output of a remote or local command and return its
    result. Raise exceptions if the command has a non-zero exitcode
    or its output is not valid JSON or is not in the expected format,
    usually ``{"local": {"return": value}}`` (+ optional keys in the "local" dict).
    """
    try:
        retcode = int(retcode)
    except (TypeError, ValueError):
        log.warning("Got an invalid retcode for host: '%s'", retcode)
        retcode = 1

    if "Permission denied" in stderr:
        # -failed to upload file- is detecting scp errors
        # Errors to ignore when Permission denied is in the stderr. For example
        # scp can get a permission denied on the target host, but they where
        # able to accurate authenticate against the box
        ignore_err = ["failed to upload file"]
        check_err = [x for x in ignore_err if stderr.count(x)]
        if not check_err:
            raise SSHPermissionDeniedError(
                stdout=stdout, stderr=stderr, retcode=retcode
            )

    result = NOT_SET
    error = None
    data = None

    try:
        data = salt.utils.json.find_json(stdout)
    except ValueError:
        # No valid JSON output was found
        error = SSHReturnDecodeError
    else:
        if isinstance(data, dict) and len(data) < 2 and "local" in data:
            result = data["local"]
            try:
                remote_retcode = result["retcode"]
            except (KeyError, TypeError):
                pass
            else:
                try:
                    # Ensure a reported local retcode is kept (at least)
                    retcode = max(retcode, remote_retcode)
                except (TypeError, ValueError):
                    log.warning(
                        "Host reported an invalid retcode: '%s'", remote_retcode
                    )
                    retcode = max(retcode, 1)

            if not isinstance(result, dict):
                # When a command has failed, the return is dumped as-is
                # without declaring it as a result, usually a string or list.
                error = SSHCommandExecutionError
            elif result_only:
                try:
                    result = result["return"]
                except KeyError:
                    error = SSHMalformedReturnError
                    result = NOT_SET
        else:
            error = SSHMalformedReturnError

    if retcode:
        error = SSHCommandExecutionError
    if error is not None:
        raise error(
            stdout=stdout,
            stderr=stderr,
            retcode=retcode,
            result=result,
            parsed=data,
        )
    return result



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
    :codeauthor: Tomas Sirny <tsirny@gmail.com>
"""

from tests.integration.cloud.helpers.cloud_test_base import TIMEOUT, CloudTest


class GCETest(CloudTest):
    """
    Integration tests for the GCE cloud provider in Salt-Cloud
    """

    PROVIDER = "gce"
    REQUIRED_PROVIDER_CONFIG_ITEMS = (
        "project",
        "service_account_email_address",
        "service_account_private_key",
    )

    def test_instance(self):
        """
        Tests creating and deleting an instance on GCE
        """

        # create the instance
        ret_str = self.run_cloud(f"-p gce-test {self.instance_name}", timeout=TIMEOUT)

        # check if instance returned with salt installed
        self.assertInstanceExists(ret_str)
        self.assertDestroyInstance()

    def test_instance_extra(self):
        """
        Tests creating and deleting an instance on GCE
        """

        # create the instance
        ret_str = self.run_cloud(
            f"-p gce-test-extra {self.instance_name}", timeout=TIMEOUT
        )

        # check if instance returned with salt installed
        self.assertInstanceExists(ret_str)
        self.assertDestroyInstance()



"""
Github User State Module

.. versionadded:: 2016.3.0

This state is used to ensure presence of users in the Organization.

.. code-block:: yaml

    ensure user test is present in github:
        github.present:
            - name: 'Example TestUser1'
            - email: example@domain.com
            - username: 'gitexample'
"""

import datetime
import logging
import time

from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if the github module is available in __salt__
    """
    if "github.list_users" in __salt__:
        return "github"
    return (False, "github module could not be loaded")


def present(name, profile="github", **kwargs):
    """
    Ensure a user is present

    .. code-block:: yaml

        ensure user test is present in github:
            github.present:
                - name: 'gitexample'

    The following parameters are required:

    name
        This is the github handle of the user in the organization
    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    target = __salt__["github.get_user"](name, profile=profile, **kwargs)

    # If the user has a valid github handle and is not in the org already
    if not target:
        ret["result"] = False
        ret["comment"] = f"Couldnt find user {name}"
    elif isinstance(target, bool) and target:
        ret["comment"] = f"User {name} is already in the org "
        ret["result"] = True
    elif (
        not target.get("in_org", False) and target.get("membership_state") != "pending"
    ):
        if __opts__["test"]:
            ret["comment"] = f"User {name} will be added to the org"
            return ret

        # add the user
        result = __salt__["github.add_user"](name, profile=profile, **kwargs)

        if result:
            ret["changes"].setdefault("old", None)
            ret["changes"].setdefault("new", f"User {name} exists in the org now")
            ret["result"] = True
        else:
            ret["result"] = False
            ret["comment"] = f"Failed to add user {name} to the org"
    else:
        ret["comment"] = f"User {name} has already been invited."
        ret["result"] = True

    return ret


def absent(name, profile="github", **kwargs):
    """
    Ensure a github user is absent

    .. code-block:: yaml

        ensure user test is absent in github:
            github.absent:
                - name: 'Example TestUser1'
                - email: example@domain.com
                - username: 'gitexample'

    The following parameters are required:

    name
        Github handle of the user in organization

    """
    email = kwargs.get("email")
    full_name = kwargs.get("fullname")

    ret = {
        "name": name,
        "changes": {},
        "result": None,
        "comment": f"User {name} is absent.",
    }

    target = __salt__["github.get_user"](name, profile=profile, **kwargs)

    if target:
        if isinstance(target, bool) or target.get("in_org", False):
            if __opts__["test"]:
                ret["comment"] = f"User {name} will be deleted"
                ret["result"] = None
                return ret

            result = __salt__["github.remove_user"](name, profile=profile, **kwargs)

            if result:
                ret["comment"] = f"Deleted user {name}"
                ret["changes"].setdefault("old", f"User {name} exists")
                ret["changes"].setdefault("new", f"User {name} deleted")
                ret["result"] = True
            else:
                ret["comment"] = f"Failed to delete {name}"
                ret["result"] = False
        else:
            ret["comment"] = f"User {name} has already been deleted!"
            ret["result"] = True
    else:
        ret["comment"] = f"User {name} does not exist"
        ret["result"] = True
        return ret

    return ret


def team_present(
    name,
    description=None,
    repo_names=None,
    privacy="secret",
    permission="pull",
    members=None,
    enforce_mfa=False,
    no_mfa_grace_seconds=0,
    profile="github",
    **kwargs,
):
    """
    Ensure a team is present

    name
        This is the name of the team in the organization.

    description
        The description of the team.

    repo_names
        The names of repositories to add the team to.

    privacy
        The level of privacy for the team, can be 'secret' or 'closed'. Defaults
        to secret.

    permission
        The default permission for new repositories added to the team, can be
        'pull', 'push' or 'admin'. Defaults to pull.

    members
        The members belonging to the team, specified as a dict of member name to
        optional configuration. Options include 'enforce_mfa_from' and 'mfa_exempt'.

    enforce_mfa
        Whether to enforce MFA requirements on members of the team. If True then
        all members without `mfa_exempt: True` configured will be removed from
        the team. Note that `no_mfa_grace_seconds` may be set to allow members
        a grace period.

    no_mfa_grace_seconds
        The number of seconds of grace time that a member will have to enable MFA
        before being removed from the team. The grace period will begin from
        `enforce_mfa_from` on the member configuration, which defaults to
        1970/01/01.

    Example:

    .. code-block:: yaml

        Ensure team test is present in github:
            github.team_present:
                - name: 'test'
                - members:
                    user1: {}
                    user2: {}

        Ensure team test_mfa is present in github:
            github.team_present:
                - name: 'test_mfa'
                - members:
                    user1:
                        enforce_mfa_from: 2016/06/15
                - enforce_mfa: True

    .. versionadded:: 2016.11.0
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    target = __salt__["github.get_team"](name, profile=profile, **kwargs)
    test_comments = []

    if target:  # Team already exists
        parameters = {}
        if description is not None and target["description"] != description:
            parameters["description"] = description
        if permission is not None and target["permission"] != permission:
            parameters["permission"] = permission
        if privacy is not None and target["privacy"] != privacy:
            parameters["privacy"] = privacy

        if len(parameters) > 0:
            if __opts__["test"]:
                test_comments.append(
                    f"Team properties are set to be edited: {parameters}"
                )
                ret["result"] = None
            else:
                result = __salt__["github.edit_team"](
                    name, profile=profile, **parameters
                )
                if result:
                    ret["changes"]["team"] = {
                        "old": f"Team properties were {target}",
                        "new": "Team properties (that changed) are {}".format(
                            parameters
                        ),
                    }
                else:
                    ret["result"] = False
                    ret["comment"] = "Failed to update team properties."
                    return ret

        manage_repos = repo_names is not None
        current_repos = set(
            __salt__["github.list_team_repos"](name, profile=profile).keys()
        )
        repo_names = set(repo_names or [])

        repos_to_add = repo_names - current_repos
        repos_to_remove = current_repos - repo_names if repo_names else []

        if repos_to_add:
            if __opts__["test"]:
                test_comments.append(
                    "Team {} will have the following repos added: {}.".format(
                        name, list(repos_to_add)
                    )
                )
                ret["result"] = None
            else:
                for repo_name in repos_to_add:
                    result = __salt__["github.add_team_repo"](
                        repo_name, name, profile=profile, **kwargs
                    )
                    if result:
                        ret["changes"][repo_name] = {
                            "old": f"Repo {repo_name} is not in team {name}",
                            "new": f"Repo {repo_name} is in team {name}",
                        }
                    else:
                        ret["result"] = False
                        ret["comment"] = "Failed to add repo {} to team {}.".format(
                            repo_name, name
                        )
                        return ret

        if repos_to_remove:
            if __opts__["test"]:
                test_comments.append(
                    "Team {} will have the following repos removed: {}.".format(
                        name, list(repos_to_remove)
                    )
                )
                ret["result"] = None
            else:
                for repo_name in repos_to_remove:
                    result = __salt__["github.remove_team_repo"](
                        repo_name, name, profile=profile, **kwargs
                    )
                    if result:
                        ret["changes"][repo_name] = {
                            "old": f"Repo {repo_name} is in team {name}",
                            "new": f"Repo {repo_name} is not in team {name}",
                        }
                    else:
                        ret["result"] = False
                        ret["comment"] = (
                            "Failed to remove repo {} from team {}.".format(
                                repo_name, name
                            )
                        )
                        return ret

    else:  # Team does not exist - it will be created.
        if __opts__["test"]:
            ret["comment"] = f"Team {name} is set to be created."
            ret["result"] = None
            return ret

        result = __salt__["github.add_team"](
            name,
            description=description,
            repo_names=repo_names,
            permission=permission,
            privacy=privacy,
            profile=profile,
            **kwargs,
        )
        if result:
            ret["changes"]["team"] = {}
            ret["changes"]["team"]["old"] = None
            ret["changes"]["team"]["new"] = f"Team {name} has been created"
        else:
            ret["result"] = False
            ret["comment"] = f"Failed to create team {name}."
            return ret

    manage_members = members is not None

    mfa_deadline = datetime.datetime.utcnow() - datetime.timedelta(
        seconds=no_mfa_grace_seconds
    )
    members_no_mfa = __salt__["github.list_members_without_mfa"](profile=profile)

    members_lower = {}
    for member_name, info in members or {}.items():
        members_lower[member_name.lower()] = info

    member_change = False
    current_members = __salt__["github.list_team_members"](name, profile=profile)

    for member, member_info in members or {}.items():
        log.info("Checking member %s in team %s", member, name)

        if member.lower() not in current_members:
            if enforce_mfa and _member_violates_mfa(
                member, member_info, mfa_deadline, members_no_mfa
            ):
                if __opts__["test"]:
                    test_comments.append(
                        "User {} will not be added to the "
                        "team because they do not have MFA."
                        "".format(member)
                    )
            else:  # Add to team
                member_change = True
                if __opts__["test"]:
                    test_comments.append(f"User {member} set to be added to the team.")
                    ret["result"] = None
                else:
                    result = __salt__["github.add_team_member"](
                        member, name, profile=profile, **kwargs
                    )
                    if result:
                        ret["changes"][member] = {}
                        ret["changes"][member][
                            "old"
                        ] = f"User {member} is not in team {name}"
                        ret["changes"][member]["new"] = "User {} is in team {}".format(
                            member, name
                        )
                    else:
                        ret["result"] = False
                        ret["comment"] = "Failed to add user {} to team {}.".format(
                            member, name
                        )
                        return ret

    for member in current_members:
        mfa_violation = False
        if member in members_lower:
            mfa_violation = _member_violates_mfa(
                member, members_lower[member], mfa_deadline, members_no_mfa
            )
        if (
            manage_members
            and member not in members_lower
            or (enforce_mfa and mfa_violation)
        ):
            # Remove from team
            member_change = True
            if __opts__["test"]:
                if mfa_violation:
                    test_comments.append(
                        "User {} set to be removed from the "
                        "team because they do not have MFA.".format(member)
                    )
                else:
                    test_comments.append(
                        f"User {member} set to be removed from the team."
                    )
                ret["result"] = None
            else:
                result = __salt__["github.remove_team_member"](
                    member, name, profile=profile, **kwargs
                )
                if result:
                    extra_changes = " due to MFA violation" if mfa_violation else ""
                    ret["changes"][member] = {
                        "old": f"User {member} is in team {name}",
                        "new": "User {} is not in team {}{}".format(
                            member, name, extra_changes
                        ),
                    }
                else:
                    ret["result"] = False
                    ret["comment"] = "Failed to remove user {} from team {}.".format(
                        member, name
                    )
                    return ret

    if member_change:  # Refresh team cache
        __salt__["github.list_team_members"](
            name, profile=profile, ignore_cache=False, **kwargs
        )

    if len(test_comments) > 0:
        ret["comment"] = "\n".join(test_comments)
    return ret


def _member_violates_mfa(member, member_info, mfa_deadline, members_without_mfa):
    if member_info.get("mfa_exempt", False):
        return False
    enforce_mfa_from = datetime.datetime.strptime(
        member_info.get("enforce_mfa_from", "1970/01/01"), "%Y/%m/%d"
    )
    return member.lower() in members_without_mfa and (mfa_deadline > enforce_mfa_from)


def team_absent(name, profile="github", **kwargs):
    """
    Ensure a team is absent.

    Example:

    .. code-block:: yaml

        ensure team test is present in github:
            github.team_absent:
                - name: 'test'


    The following parameters are required:

    name
        This is the name of the team in the organization.

    .. versionadded:: 2016.11.0
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    target = __salt__["github.get_team"](name, profile=profile, **kwargs)

    if not target:
        ret["comment"] = f"Team {name} does not exist"
        ret["result"] = True
        return ret
    else:
        if __opts__["test"]:
            ret["comment"] = f"Team {name} will be deleted"
            ret["result"] = None
            return ret

        result = __salt__["github.remove_team"](name, profile=profile, **kwargs)

        if result:
            ret["comment"] = f"Deleted team {name}"
            ret["changes"].setdefault("old", f"Team {name} exists")
            ret["changes"].setdefault("new", f"Team {name} deleted")
            ret["result"] = True
        else:
            ret["comment"] = f"Failed to delete {name}"
            ret["result"] = False
    return ret


def repo_present(
    name,
    description=None,
    homepage=None,
    private=None,
    has_issues=None,
    has_wiki=None,
    has_downloads=None,
    auto_init=False,
    gitignore_template=None,
    license_template=None,
    teams=None,
    profile="github",
    **kwargs,
):
    """
    Ensure a repository is present

    name
        This is the name of the repository.

    description
        The description of the repository.

    homepage
        The URL with more information about the repository.

    private
        The visiblity of the repository. Note that private repositories require
        a paid GitHub account.

    has_issues
        Whether to enable issues for this repository.

    has_wiki
        Whether to enable the wiki for this repository.

    has_downloads
        Whether to enable downloads for this repository.

    auto_init
        Whether to create an initial commit with an empty README.

    gitignore_template
        The desired language or platform for a .gitignore, e.g "Haskell".

    license_template
        The desired LICENSE template to apply, e.g "mit" or "mozilla".

    teams
        The teams for which this repo should belong to, specified as a dict of
        team name to permission ('pull', 'push' or 'admin').

        .. versionadded:: 2017.7.0

    Example:

    .. code-block:: yaml

        Ensure repo my-repo is present in github:
            github.repo_present:
                - name: 'my-repo'
                - description: 'My very important repository'

    .. versionadded:: 2016.11.0
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # This is an optimization to cache all repos in the organization up front.
    # The first use of this state will collect all of the repos and save a bunch
    # of API calls for future use.
    __salt__["github.list_repos"](profile=profile)

    try:
        target = __salt__["github.get_repo_info"](name, profile=profile, **kwargs)
    except CommandExecutionError:
        target = None

    given_params = {
        "description": description,
        "homepage": homepage,
        "private": private,
        "has_issues": has_issues,
        "has_wiki": has_wiki,
        "has_downloads": has_downloads,
        "auto_init": auto_init,
        "gitignore_template": gitignore_template,
        "license_template": license_template,
    }

    # Keep track of current_teams if we've fetched them after creating a new repo
    current_teams = None

    if target:  # Repo already exists
        # Some params are only valid on repo creation
        ignore_params = ["auto_init", "gitignore_template", "license_template"]
        parameters = {}
        old_parameters = {}
        for param_name, param_value in given_params.items():
            if (
                param_value is not None
                and param_name not in ignore_params
                and target[param_name] is not param_value
                and target[param_name] != param_value
            ):
                parameters[param_name] = param_value
                old_parameters[param_name] = target[param_name]

        if len(parameters) > 0:
            repo_change = {
                "old": f"Repo properties were {old_parameters}",
                "new": f"Repo properties (that changed) are {parameters}",
            }
            if __opts__["test"]:
                ret["changes"]["repo"] = repo_change
                ret["result"] = None
            else:
                result = __salt__["github.edit_repo"](
                    name, profile=profile, **parameters
                )
                if result:
                    ret["changes"]["repo"] = repo_change
                else:
                    ret["result"] = False
                    ret["comment"] = "Failed to update repo properties."
                    return ret

    else:  # Repo does not exist - it will be created.
        repo_change = {"old": None, "new": f"Repo {name} has been created"}
        if __opts__["test"]:
            ret["changes"]["repo"] = repo_change
            ret["result"] = None
        else:
            add_params = dict(given_params)
            add_params.update(kwargs)
            result = __salt__["github.add_repo"](name, **add_params)

            if not result:
                ret["result"] = False
                ret["comment"] = f"Failed to create repo {name}."
                return ret

            # Turns out that trying to fetch teams for a new repo can 404 immediately
            # after repo creation, so this waits until we can fetch teams successfully
            # before continuing.
            for attempt in range(3):
                time.sleep(1)
                try:
                    current_teams = __salt__["github.get_repo_teams"](
                        name, profile=profile, **kwargs
                    )
                    break
                except CommandExecutionError as e:
                    log.info("Attempt %s to fetch new repo %s failed", attempt, name)

            if current_teams is None:
                ret["result"] = False
                ret["comment"] = f"Failed to verify repo {name} after creation."
                return ret

            ret["changes"]["repo"] = repo_change

    if teams is not None:
        if __opts__["test"] and not target:
            # Assume no teams if we're in test mode and the repo doesn't exist
            current_teams = []
        elif current_teams is None:
            current_teams = __salt__["github.get_repo_teams"](name, profile=profile)
        current_team_names = {t["name"] for t in current_teams}

        # First remove any teams that aren't present
        for team_name in current_team_names:
            if team_name not in teams:
                team_change = {
                    "old": f"Repo {name} is in team {team_name}",
                    "new": f"Repo {name} is not in team {team_name}",
                }

                if __opts__["test"]:
                    ret["changes"][team_name] = team_change
                    ret["result"] = None
                else:
                    result = __salt__["github.remove_team_repo"](
                        name, team_name, profile=profile
                    )
                    if result:
                        ret["changes"][team_name] = team_change
                    else:
                        ret["result"] = False
                        ret["comment"] = (
                            "Failed to remove repo {} from team {}.".format(
                                name, team_name
                            )
                        )
                        return ret

        # Next add or modify any necessary teams
        for team_name, permission in teams.items():
            if team_name not in current_team_names:  # Need to add repo to team
                team_change = {
                    "old": f"Repo {name} is not in team {team_name}",
                    "new": f"Repo {name} is in team {team_name}",
                }
                if __opts__["test"]:
                    ret["changes"][team_name] = team_change
                    ret["result"] = None
                else:
                    result = __salt__["github.add_team_repo"](
                        name, team_name, profile=profile, permission=permission
                    )
                    if result:
                        ret["changes"][team_name] = team_change
                    else:
                        ret["result"] = False
                        ret["comment"] = (
                            "Failed to remove repo {} from team {}.".format(
                                name, team_name
                            )
                        )
                        return ret
            else:
                current_permission = (
                    __salt__["github.list_team_repos"](team_name, profile=profile)
                    .get(name.lower(), {})
                    .get("permission")
                )
                if not current_permission:
                    ret["result"] = False
                    ret["comment"] = (
                        "Failed to determine current permission for team "
                        "{} in repo {}".format(team_name, name)
                    )
                    return ret
                elif current_permission != permission:
                    team_change = {
                        "old": "Repo {} in team {} has permission {}".format(
                            name, team_name, current_permission
                        ),
                        "new": "Repo {} in team {} has permission {}".format(
                            name, team_name, permission
                        ),
                    }
                    if __opts__["test"]:
                        ret["changes"][team_name] = team_change
                        ret["result"] = None
                    else:
                        result = __salt__["github.add_team_repo"](
                            name, team_name, profile=profile, permission=permission
                        )
                        if result:
                            ret["changes"][team_name] = team_change
                        else:
                            ret["result"] = False
                            ret["comment"] = (
                                "Failed to set permission on repo {} from "
                                "team {} to {}.".format(name, team_name, permission)
                            )
                            return ret
    return ret


def repo_absent(name, profile="github", **kwargs):
    """
    Ensure a repo is absent.

    Example:

    .. code-block:: yaml

        ensure repo test is absent in github:
            github.repo_absent:
                - name: 'test'

    The following parameters are required:

    name
        This is the name of the repository in the organization.

    .. versionadded:: 2016.11.0
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    try:
        target = __salt__["github.get_repo_info"](name, profile=profile, **kwargs)
    except CommandExecutionError:
        target = None

    if not target:
        ret["comment"] = f"Repo {name} does not exist"
        ret["result"] = True
        return ret
    else:
        if __opts__["test"]:
            ret["comment"] = f"Repo {name} will be deleted"
            ret["result"] = None
            return ret

        result = __salt__["github.remove_repo"](name, profile=profile, **kwargs)

        if result:
            ret["comment"] = f"Deleted repo {name}"
            ret["changes"].setdefault("old", f"Repo {name} exists")
            ret["changes"].setdefault("new", f"Repo {name} deleted")
            ret["result"] = True
        else:
            ret["comment"] = (
                "Failed to delete repo {}. Ensure the delete_repo "
                "scope is enabled if using OAuth.".format(name)
            )
            ret["result"] = False
    return ret



import textwrap

import pytest
from saltfactories.utils.functional import StateResult

pytestmark = [
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def reset_pillar(salt_call_cli):
    try:
        # Run tests
        yield
    finally:
        # Refresh pillar once all tests are done.
        ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True


@pytest.fixture
def testfile_path(tmp_path, base_env_state_tree_root_dir):
    testfile = tmp_path / "testfile"
    sls_contents = textwrap.dedent(
        """
        {}:
          file:
            - managed
            - source: salt://testfile
            - makedirs: true
            - mode: 644
        """.format(
            testfile
        )
    )
    with pytest.helpers.temp_file(
        "sls-id-test.sls", sls_contents, base_env_state_tree_root_dir
    ):
        yield testfile


@pytest.mark.usefixtures("testfile_path", "reset_pillar")
def test_state_apply_aborts_on_pillar_error(
    salt_cli,
    salt_minion,
    base_env_pillar_tree_root_dir,
):
    """
    Test state.apply with error in pillar.
    """
    pillar_top_file = textwrap.dedent(
        """
        base:
          '{}':
            - basic
        """.format(
            salt_minion.id
        )
    )
    basic_pillar_file = textwrap.dedent(
        """
        syntax_error
        """
    )

    with pytest.helpers.temp_file(
        "top.sls", pillar_top_file, base_env_pillar_tree_root_dir
    ), pytest.helpers.temp_file(
        "basic.sls", basic_pillar_file, base_env_pillar_tree_root_dir
    ):
        expected_comment = [
            "Pillar failed to render with the following messages:",
            "SLS 'basic' does not render to a dictionary",
        ]
        shell_result = salt_cli.run(
            "state.apply", "sls-id-test", minion_tgt=salt_minion.id
        )
        assert shell_result.returncode == 1
        assert shell_result.data == expected_comment


@pytest.mark.usefixtures("testfile_path", "reset_pillar")
def test_state_apply_continues_after_pillar_error_is_fixed(
    salt_cli,
    salt_minion,
    base_env_pillar_tree_root_dir,
):
    """
    Test state.apply with error in pillar.
    """
    pillar_top_file = textwrap.dedent(
        """
        base:
          '{}':
            - basic
        """
    ).format(salt_minion.id)
    basic_pillar_file_error = textwrap.dedent(
        """
        syntax_error
        """
    )
    basic_pillar_file = textwrap.dedent(
        """
        syntax_error: Fixed!
        """
    )

    # save pillar render error in minion's in-memory pillar
    with pytest.helpers.temp_file(
        "top.sls", pillar_top_file, base_env_pillar_tree_root_dir
    ), pytest.helpers.temp_file(
        "basic.sls", basic_pillar_file_error, base_env_pillar_tree_root_dir
    ):
        shell_result = salt_cli.run(
            "saltutil.refresh_pillar", minion_tgt=salt_minion.id
        )
        assert shell_result.returncode == 0

    # run state.apply with fixed pillar render error
    with pytest.helpers.temp_file(
        "top.sls", pillar_top_file, base_env_pillar_tree_root_dir
    ), pytest.helpers.temp_file(
        "basic.sls", basic_pillar_file, base_env_pillar_tree_root_dir
    ):
        shell_result = salt_cli.run(
            "state.apply", "sls-id-test", minion_tgt=salt_minion.id
        )
        assert shell_result.returncode == 0
        state_result = StateResult(shell_result.data)
        assert state_result.result is True
        assert state_result.changes == {"diff": "New file", "mode": "0644"}



#!/usr/bin/env python
'''
Welcome to the Salt repl which exposes the execution environment of a minion in
a pre-configured Python shell

__opts__, __salt__, __grains__, and __pillar__ are available.

Jinja can be tested with full access to the above structures in the usual way:

    JINJA("""\\
        I am {{ salt['cmd.run']('whoami') }}.

        {% if otherstuff %}
        Some other stuff here
        {% endif %}
    """, otherstuff=True)

A history file is maintained in ~/.saltsh_history.
completion behavior can be customized via the ~/.inputrc file.

'''
# pylint: disable=file-perms


import atexit
import builtins
import os
import pprint  # pylint: disable=unused-import
import readline
import sys
from code import InteractiveConsole

import jinja2

import salt.client
import salt.config
import salt.loader
import salt.output
import salt.pillar
import salt.runner

# pylint: disable=unused-import
# These are imported to be available in the spawned shell
import salt.utils.yaml

# pylint: enable=unused-import

HISTFILE = "{HOME}/.saltsh_history".format(**os.environ)


def savehist():
    """
    Save the history file
    """
    readline.write_history_file(HISTFILE)


def get_salt_vars():
    """
    Return all the Salt-usual double-under data structures for a minion
    """
    # Create the Salt __opts__ variable
    __opts__ = salt.config.client_config(
        os.environ.get("SALT_MINION_CONFIG", "/etc/salt/minion")
    )

    # Populate grains if it hasn't been done already
    if "grains" not in __opts__ or not __opts__["grains"]:
        __opts__["grains"] = salt.loader.grains(__opts__)

    # file_roots and pillar_roots should be set in the minion config
    if "file_client" not in __opts__ or not __opts__["file_client"]:
        __opts__["file_client"] = "local"

    # ensure we have a minion id
    if "id" not in __opts__ or not __opts__["id"]:
        __opts__["id"] = "saltsh_mid"

    # Populate template variables
    __salt__ = salt.loader.minion_mods(__opts__)
    __grains__ = __opts__["grains"]

    if __opts__["file_client"] == "local":
        __pillar__ = salt.pillar.get_pillar(
            __opts__,
            __grains__,
            __opts__.get("id"),
            __opts__.get("saltenv"),
        ).compile_pillar()
    else:
        __pillar__ = {}

    # pylint: disable=invalid-name,unused-variable,possibly-unused-variable
    def JINJA(x, **y):
        return jinja2.Template(x).render(
            grains=__grains__, salt=__salt__, opts=__opts__, pillar=__pillar__, **y
        )

    # pylint: enable=invalid-name,unused-variable,possibly-unused-variable

    return locals()


def main():
    """
    The main entry point
    """
    salt_vars = get_salt_vars()

    def salt_outputter(value):
        """
        Use Salt's outputters to print values to the shell
        """
        if value is not None:
            builtins._ = value
            salt.output.display_output(value, "", salt_vars["__opts__"])

    sys.displayhook = salt_outputter

    # Set maximum number of items that will be written to the history file
    readline.set_history_length(300)

    if os.path.exists(HISTFILE):
        readline.read_history_file(HISTFILE)

    atexit.register(savehist)
    atexit.register(lambda: sys.stdout.write("Salt you later!\n"))

    saltrepl = InteractiveConsole(locals=salt_vars)
    saltrepl.interact(banner=__doc__)


if __name__ == "__main__":
    main()



import warnings

from salt.utils.decorators.extension_deprecation import extension_deprecation_message


@extension_deprecation_message(3009, "salt_mod", "http://www.example.com")
def salt_func():
    return True


@extension_deprecation_message(3009, "salt_mod", "http://www.example.com")
def get_configured_provider():
    return True


def test_extension_deprecation():
    """
    this tests the extension_deprecation_message decorator to
    ensure that the warning is raised when a decorated function
    is called.
    """
    expected_deprecation_message = (
        "The 'salt_mod' functionality in Salt has been deprecated and "
        "its functionality will be removed in version 3009.0 (Potassium) "
        "in favor of the saltext.salt_mod Salt Extension. (http://www.example.com)"
    )
    with warnings.catch_warnings(record=True) as catch_warnings:
        ret = salt_func()
        assert ret
        assert len(catch_warnings) == 1
        assert issubclass(catch_warnings[-1].category, DeprecationWarning)
        assert str(catch_warnings[-1].message) == expected_deprecation_message

    with warnings.catch_warnings(record=True) as catch_warnings:
        ret = get_configured_provider()
        assert ret
        assert len(catch_warnings) == 0



import hashlib
from textwrap import dedent

import pytest

import salt.modules.pdbedit as pdbedit
from tests.support.mock import MagicMock, patch

try:
    hashlib.new("md4", "".encode("utf-16le"))
    MD4_SUPPORTED = True
except ValueError:
    MD4_SUPPORTED = False


@pytest.fixture
def configure_loader_modules():
    return {pdbedit: {}}


@pytest.mark.parametrize("verbose", [True, False])
def test_when_no_users_returned_no_data_should_be_returned(verbose):
    expected_users = {} if verbose else []
    with patch.dict(
        pdbedit.__salt__,
        {
            "cmd.run_all": MagicMock(
                return_value={"stdout": "", "stderr": "", "retcode": 0}
            )
        },
    ):
        actual_users = pdbedit.list_users(verbose=verbose)

    assert actual_users == expected_users


def test_when_verbose_and_retcode_is_nonzero_output_should_be_had():
    expected_stderr = "this is something fnord"
    with patch.dict(
        pdbedit.__salt__,
        {
            "cmd.run_all": MagicMock(
                return_value={"stdout": "", "stderr": expected_stderr, "retcode": 1}
            )
        },
    ), patch("salt.modules.pdbedit.log.error", autospec=True) as fake_error_log:
        pdbedit.list_users(verbose=True)
        actual_error = fake_error_log.mock_calls[0].args[0]
        assert actual_error == expected_stderr


def test_when_verbose_and_single_good_output_expected_data_should_be_parsed():
    expected_data = {
        "roscivs": {
            "unix username": "roscivs",
            "nt username": "bottia",
            "full name": "Roscivs Bottia",
            "user sid": "42",
            "primary group sid": "99",
            "home directory": r"\\samba\roscivs",
            "account desc": "separators! xxx so long and thanks for all the fish",
            "logoff time": "Sat, 14 Aug 2010 15:06:39 UTC",
            "kickoff time": "Sat, 14 Aug 2010 15:06:39 UTC",
            "password must change": "never",
        }
    }
    pdb_output = dedent(
        r"""
        Unix username:        roscivs
        NT username:          bottia
        User SID:             42
        Primary Group SID:    99
        Full Name:            Roscivs Bottia
        Home Directory:       \\samba\roscivs
        Account desc:         separators! xxx so long and thanks for all the fish
        Logoff time:          Sat, 14 Aug 2010 15:06:39 UTC
        Kickoff time:         Sat, 14 Aug 2010 15:06:39 UTC
        Password must change: never
        """
    ).strip()

    with patch.dict(
        pdbedit.__salt__,
        {
            "cmd.run_all": MagicMock(
                return_value={"stdout": pdb_output, "stderr": "", "retcode": 0}
            )
        },
    ):
        actual_data = pdbedit.list_users(verbose=True)
        assert actual_data == expected_data


def test_when_verbose_and_multiple_records_present_data_should_be_correctly_parsed():
    expected_data = {
        "roscivs": {
            "unix username": "roscivs",
            "nt username": "bottia",
            "user sid": "42",
        },
        "srilyk": {
            "unix username": "srilyk",
            "nt username": "srilyk",
            "account desc": "trololollol",
            "user sid": "99",
        },
        "jewlz": {
            "unix username": "jewlz",
            "nt username": "flutterbies",
            "user sid": "4",
        },
    }
    pdb_output = dedent(
        """
        -------------
        Unix username:        roscivs
        NT username:          bottia
        User SID:             42
        -------------
        Unix username:        srilyk
        NT username:          srilyk
        User SID:             99
        Account desc:         trololol\x1dlol
        -------------
        Unix username:        jewlz
        NT username:          flutterbies
        User SID:             4
        -------------
        -------------
        -------------
        """
    ).strip()

    with patch.dict(
        pdbedit.__salt__,
        {
            "cmd.run_all": MagicMock(
                return_value={"stdout": pdb_output, "stderr": "", "retcode": 0}
            )
        },
    ):
        actual_data = pdbedit.list_users(verbose=True)
        assert actual_data == expected_data


@pytest.mark.skipif(not MD4_SUPPORTED, reason="Requires md4")
def test_create_with_existing_user_updates_password():
    with patch(
        "salt.modules.pdbedit.list_users", MagicMock(return_value=["Foo"])
    ), patch(
        "salt.modules.pdbedit.get_user",
        MagicMock(return_value={"nt hash": "old value"}),
    ), patch.dict(
        pdbedit.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
    ):
        ret = pdbedit.create("Foo", "secret")
        assert {"Foo": "updated"} == ret



import pathlib

import pytest

from tests.support.runtests import RUNTIME_VARS


@pytest.fixture(scope="session", autouse=True)
def _create_old_tempdir():
    pathlib.Path(RUNTIME_VARS.TMP).mkdir(exist_ok=True, parents=True)



"""
Example beacon to use with salt-proxy

.. code-block:: yaml

    beacons:
      proxy_example:
        endpoint: beacon
"""

import logging

import salt.utils.beacons
import salt.utils.http

# Important: If used with salt-proxy
# this is required for the beacon to load!!!
__proxyenabled__ = ["*"]

__virtualname__ = "proxy_example"

log = logging.getLogger(__name__)


def __virtual__():
    """
    Trivially let the beacon load for the test example.
    For a production beacon we should probably have some expression here.
    """
    return True


def validate(config):
    """
    Validate the beacon configuration
    """
    if not isinstance(config, list):
        return False, "Configuration for proxy_example beacon must be a list."
    return True, "Valid beacon configuration"


def beacon(config):
    """
    Called several times each second
    https://docs.saltproject.io/en/latest/topics/beacons/#the-beacon-function

    .. code-block:: yaml

        beacons:
          proxy_example:
            - endpoint: beacon
    """
    # Important!!!
    # Although this toy example makes an HTTP call
    # to get beacon information
    # please be advised that doing CPU or IO intensive
    # operations in this method will cause the beacon loop
    # to block.
    config = salt.utils.beacons.list_to_dict(config)

    beacon_url = "{}{}".format(__opts__["proxy"]["url"], config["endpoint"])
    ret = salt.utils.http.query(beacon_url, decode_type="json", decode=True)
    return [ret["dict"]]



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>

    TestCase for the salt.modules.at module
"""

import pytest

import salt.modules.at as at
import salt.utils.path
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {at: {}}


@pytest.fixture
def atq_output():
    return {
        "jobs": [
            {
                "date": "2014-12-11",
                "job": 101,
                "queue": "A",
                "tag": "",
                "time": "19:48:47",
                "user": "B",
            }
        ]
    }


def test_atq_not_available():
    """
    Tests the at.atq not available for any type of os_family.
    """
    with patch("salt.modules.at._cmd", MagicMock(return_value=None)):
        with patch.dict(at.__grains__, {"os_family": "RedHat"}):
            assert at.atq() == "'at.atq' is not available."

        with patch.dict(at.__grains__, {"os_family": ""}):
            assert at.atq() == "'at.atq' is not available."


def test_atq_no_jobs_available():
    """
    Tests the no jobs available for any type of os_family.
    """
    with patch("salt.modules.at._cmd", MagicMock(return_value="")):
        with patch.dict(at.__grains__, {"os_family": "RedHat"}):
            assert at.atq() == {"jobs": []}

        with patch.dict(at.__grains__, {"os_family": ""}):
            assert at.atq() == {"jobs": []}


def test_atq_list():
    """
    Tests the list all queued and running jobs.
    """
    with patch("salt.modules.at._cmd") as salt_modules_at__cmd_mock:
        salt_modules_at__cmd_mock.return_value = (
            "101\tThu Dec 11             19:48:47 2014 A B"
        )
        with patch.dict(at.__grains__, {"os_family": "", "os": ""}):
            assert at.atq() == {
                "jobs": [
                    {
                        "date": "2014-12-11",
                        "job": 101,
                        "queue": "A",
                        "tag": "",
                        "time": "19:48:00",
                        "user": "B",
                    }
                ]
            }

        salt_modules_at__cmd_mock.return_value = (
            "101\t2014-12-11             19:48:47 A B"
        )
        with patch.dict(at.__grains__, {"os_family": "RedHat", "os": ""}):
            assert at.atq() == {
                "jobs": [
                    {
                        "date": "2014-12-11",
                        "job": 101,
                        "queue": "A",
                        "tag": "",
                        "time": "19:48:47",
                        "user": "B",
                    }
                ]
            }

        salt_modules_at__cmd_mock.return_value = (
            "SALT: Dec 11,             2014 19:48 A 101 B"
        )
        with patch.dict(at.__grains__, {"os_family": "", "os": "OpenBSD"}):
            assert at.atq() == {
                "jobs": [
                    {
                        "date": "2014-12-11",
                        "job": "101",
                        "queue": "B",
                        "tag": "",
                        "time": "19:48:00",
                        "user": "A",
                    }
                ]
            }


def test_atrm(atq_output):
    """
    Tests for remove jobs from the queue.
    """
    with patch("salt.modules.at.atq", MagicMock(return_value=atq_output)):
        with patch.object(salt.utils.path, "which", return_value=None):
            assert at.atrm() == "'at.atrm' is not available."

        with patch.object(salt.utils.path, "which", return_value=True):
            assert at.atrm() == {"jobs": {"removed": [], "tag": None}}

        with patch.object(at, "_cmd", return_value=True):
            with patch.object(salt.utils.path, "which", return_value=True):
                assert at.atrm("all") == {"jobs": {"removed": ["101"], "tag": None}}

        with patch.object(at, "_cmd", return_value=True):
            with patch.object(salt.utils.path, "which", return_value=True):
                assert at.atrm(101) == {"jobs": {"removed": ["101"], "tag": None}}

        with patch.object(at, "_cmd", return_value=None):
            assert at.atrm(101) == "'at.atrm' is not available."


def test_jobcheck(atq_output):
    """
    Tests for check the job from queue.
    """
    with patch("salt.modules.at.atq", MagicMock(return_value=atq_output)):
        assert at.jobcheck() == {"error": "You have given a condition"}

        assert at.jobcheck(runas="foo") == {
            "note": "No match jobs or time format error",
            "jobs": [],
        }

        assert at.jobcheck(
            runas="B", tag="", hour=19, minute=48, day=11, month=12, Year=2014
        ) == {
            "jobs": [
                {
                    "date": "2014-12-11",
                    "job": 101,
                    "queue": "A",
                    "tag": "",
                    "time": "19:48:47",
                    "user": "B",
                }
            ]
        }


def test_at(atq_output):
    """
    Tests for add a job to the queue.
    """
    with patch("salt.modules.at.atq", MagicMock(return_value=atq_output)):
        assert at.at() == {"jobs": []}

        with patch.object(salt.utils.path, "which", return_value=None):
            assert (
                at.at("12:05am", "/sbin/reboot", tag="reboot")
                == "'at.at' is not available."
            )

        with patch.object(salt.utils.path, "which", return_value=True):
            with patch.dict(at.__grains__, {"os_family": "RedHat"}):
                mock = MagicMock(return_value=None)
                with patch.dict(at.__salt__, {"cmd.run": mock}):
                    assert (
                        at.at("12:05am", "/sbin/reboot", tag="reboot")
                        == "'at.at' is not available."
                    )

                mock = MagicMock(return_value="Garbled time")
                with patch.dict(at.__salt__, {"cmd.run": mock}):
                    assert at.at("12:05am", "/sbin/reboot", tag="reboot") == {
                        "jobs": [],
                        "error": "invalid timespec",
                    }

                mock = MagicMock(return_value="warning: commands\nA B")
                with patch.dict(at.__salt__, {"cmd.run": mock}):
                    with patch.dict(at.__grains__, {"os": "OpenBSD"}):
                        assert at.at("12:05am", "/sbin/reboot", tag="reboot") == {
                            "jobs": [
                                {
                                    "date": "2014-12-11",
                                    "job": 101,
                                    "queue": "A",
                                    "tag": "",
                                    "time": "19:48:47",
                                    "user": "B",
                                }
                            ]
                        }

            with patch.dict(at.__grains__, {"os_family": ""}):
                mock = MagicMock(return_value=None)
                with patch.dict(at.__salt__, {"cmd.run": mock}):
                    assert (
                        at.at("12:05am", "/sbin/reboot", tag="reboot")
                        == "'at.at' is not available."
                    )


def test_atc():
    """
    Tests for atc
    """
    with patch.object(at, "_cmd", return_value=None):
        assert at.atc(101) == "'at.atc' is not available."

    with patch.object(at, "_cmd", return_value=""):
        assert at.atc(101) == {"error": "invalid job id '101'"}

    with patch.object(at, "_cmd", return_value="101\tThu Dec 11 19:48:47 2014 A B"):
        assert at.atc(101) == "101\tThu Dec 11 19:48:47 2014 A B"



import pytest

import salt.modules.win_certutil as certutil
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {certutil: {}}


def test_get_serial():
    """
    Test getting the serial number from a certificate
    """
    expected = "180720d39cd2db3244ba037417241e90"
    mock = MagicMock(
        return_value=(
            "CertInfo\r\n"
            "Cert Serial Number: 180720d39cd2db3244ba037417241e90\r\n"
            "\r\n"
            "OtherStuff"
        )
    )
    with patch.dict(certutil.__salt__, {"cmd.run": mock}), patch.dict(
        certutil.__salt__,
        {"cp.cache_file": MagicMock(return_value="/path/to/cert.cer")},
    ), patch("os.path.exists", MagicMock(return_value=True)):
        out = certutil.get_cert_serial("/path/to/cert.cer")
        mock.assert_called_once_with('certutil.exe -silent -verify "/path/to/cert.cer"')
        assert expected == out


def test_get_serials():
    """
    Test getting all the serial numbers from a store
    """
    expected = [
        "180720d39cd2db3244ba037417241e90",
        "1768ac4e5b72bf1d0df0df118b34b959",
    ]
    mock = MagicMock(
        return_value=(
            "CertInfo\r\n"
            "================ Certificate 0 ================\r\n"
            "Serial Number: 180720d39cd2db3244ba037417241e90\r\n"
            "OtherStuff\r\n"
            "\r\n"
            "================ Certificate 1 ================\r\n"
            "Serial Number: 1768ac4e5b72bf1d0df0df118b34b959\r\n"
            "OtherStuff"
        )
    )
    with patch.dict(certutil.__salt__, {"cmd.run": mock}):
        out = certutil.get_stored_cert_serials("TrustedPublisher")
        mock.assert_called_once_with('certutil.exe -store "TrustedPublisher"')
        assert expected == out


def test_add_store():
    """
    Test adding a certificate to a specific store
    """
    cmd_mock = MagicMock(
        return_value=(
            "CertInfo\r\n"
            "================ Certificate 0 ================\r\n"
            "Serial Number: 180720d39cd2db3244ba037417241e90\r\n"
            "OtherStuff"
        )
    )
    cache_mock = MagicMock(return_value="/tmp/cert.cer")
    with patch.dict(
        certutil.__salt__, {"cmd.run": cmd_mock, "cp.cache_file": cache_mock}
    ), patch("os.path.exists", MagicMock(return_value=True)):
        certutil.add_store("salt://path/to/file", "TrustedPublisher")
        cmd_mock.assert_called_once_with(
            'certutil.exe -addstore TrustedPublisher "/tmp/cert.cer"'
        )
        cache_mock.assert_called_once_with("salt://path/to/file", "base")


def test_del_store():
    """
    Test removing a certificate to a specific store
    """
    with patch("salt.modules.win_certutil.get_cert_serial") as cert_serial_mock:
        cmd_mock = MagicMock(
            return_value=(
                "CertInfo\r\n"
                "================ Certificate 0 ================\r\n"
                "Serial Number: 180720d39cd2db3244ba037417241e90\r\n"
                "OtherStuff"
            )
        )
        cache_mock = MagicMock(return_value="/tmp/cert.cer")
        cert_serial_mock.return_value = "ABCDEF"
        with patch.dict(
            certutil.__salt__, {"cmd.run": cmd_mock, "cp.cache_file": cache_mock}
        ), patch("os.path.exists", MagicMock(return_value=True)):
            certutil.del_store("salt://path/to/file", "TrustedPublisher")
            cmd_mock.assert_called_once_with(
                'certutil.exe -delstore TrustedPublisher "ABCDEF"'
            )
            cache_mock.assert_called_once_with("salt://path/to/file", "base")



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_prepend_issue_27401_makedirs(file, tmp_path):
    """
    file.prepend but create directories if needed as an option, and create
    the file if it doesn't exist
    """
    fname = "prepend_issue_27401"
    name = tmp_path / fname

    # Non existing file get's touched
    ret = file.prepend(name=str(name), text="cheese", makedirs=True)
    assert ret.result is True
    assert name.is_file()
    assert name.read_text() == "cheese\n"

    # Nested directory and file get's touched
    name = tmp_path / "issue_27401" / fname
    ret = file.prepend(name=str(name), text="cheese", makedirs=True)
    assert ret.result is True
    assert name.is_file()
    assert name.read_text() == "cheese\n"
    assert name.parent.is_dir()

    # Parent directory exists but file does not and makedirs is False
    name = name.with_name(fname + "2")
    ret = file.prepend(name=str(name), text="cheese", makedirs=False)
    assert ret.result is True
    assert name.is_file()
    assert name.read_text() == "cheese\n"
    assert name.parent.is_dir()



import os
import time

import tornado.gen

import salt.transport.tcp


async def test_pub_channel(master_opts, minion_opts, io_loop):
    def presence_callback(client):
        pass

    def remove_presence_callback(client):
        pass

    master_opts["transport"] = "tcp"
    minion_opts.update(master_ip="127.0.0.1", transport="tcp")

    server = salt.transport.tcp.TCPPublishServer(
        master_opts,
        pub_host="127.0.0.1",
        pub_port=master_opts["publish_port"],
        pull_path=os.path.join(master_opts["sock_dir"], "publish_pull.ipc"),
    )

    client = salt.transport.tcp.TCPPubClient(
        minion_opts,
        io_loop,
        host="127.0.0.1",
        port=master_opts["publish_port"],
    )

    payloads = []

    publishes = []

    async def publish_payload(payload, callback):
        await server.publish_payload(payload)
        payloads.append(payload)

    async def on_recv(message):
        publishes.append(message)

    io_loop.add_callback(
        server.publisher, publish_payload, presence_callback, remove_presence_callback
    )

    # Wait for socket to bind.
    await tornado.gen.sleep(3)

    await client.connect(master_opts["publish_port"])
    client.on_recv(on_recv)

    await server.publish({"meh": "bah"})

    start = time.monotonic()
    try:
        while not publishes:
            await tornado.gen.sleep(0.3)
            if time.monotonic() - start > 30:
                assert False, "Message not published after 30 seconds"
    finally:
        server.close()
        server.pub_server.close()
        client.close()



import pytest
from tornado.httpclient import HTTPError


@pytest.fixture
def app(app):
    app.wsgi_application.config["global"]["tools.hypermedia_out.on"] = True
    return app


async def test_default_accept(http_client, content_type_map):
    response = await http_client.fetch("/", method="GET")
    assert response.headers["Content-Type"] == content_type_map["json"]


async def test_unsupported_accept(http_client):
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/", method="GET", headers={"Accept": "application/ms-word"}
        )
    assert exc.value.code == 406


async def test_json_out(http_client, content_type_map):
    response = await http_client.fetch(
        "/", method="GET", headers={"Accept": content_type_map["json"]}
    )
    assert response.headers["Content-Type"] == content_type_map["json"]


async def test_yaml_out(http_client, content_type_map):
    response = await http_client.fetch(
        "/", method="GET", headers={"Accept": content_type_map["yaml"]}
    )
    assert response.headers["Content-Type"] == content_type_map["yaml"]



"""
XML file manager

.. versionadded:: 3000
"""

import logging
import xml.etree.ElementTree as ET

log = logging.getLogger(__name__)


# Define the module's virtual name
__virtualname__ = "xml"


def __virtual__():
    """
    Only load the module if all modules are imported correctly.
    """
    return __virtualname__


def get_value(file, element):
    """
    Returns the value of the matched xpath element

    CLI Example:

    .. code-block:: bash

        salt '*' xml.get_value /tmp/test.xml ".//element"
    """
    try:
        root = ET.parse(file)
        element = root.find(element)
        return element.text
    except AttributeError:
        log.error("Unable to find element matching %s", element)
        return False


def set_value(file, element, value):
    """
    Sets the value of the matched xpath element

    CLI Example:

    .. code-block:: bash

        salt '*' xml.set_value /tmp/test.xml ".//element" "new value"
    """
    try:
        root = ET.parse(file)
        relement = root.find(element)
    except AttributeError:
        log.error("Unable to find element matching %s", element)
        return False
    relement.text = str(value)
    root.write(file)
    return True


def get_attribute(file, element):
    """
    Return the attributes of the matched xpath element.

    CLI Example:

    .. code-block:: bash

        salt '*' xml.get_attribute /tmp/test.xml ".//element[@id='3']"
    """
    try:
        root = ET.parse(file)
        element = root.find(element)
        return element.attrib
    except AttributeError:
        log.error("Unable to find element matching %s", element)
        return False


def set_attribute(file, element, key, value):
    """
    Set the requested attribute key and value for matched xpath element.

    CLI Example:

    .. code-block:: bash

        salt '*' xml.set_attribute /tmp/test.xml ".//element[@id='3']" editedby "gal"
    """
    try:
        root = ET.parse(file)
        element = root.find(element)
    except AttributeError:
        log.error("Unable to find element matching %s", element)
        return False
    element.set(key, str(value))
    root.write(file)
    return True



"""
Redis SDB module
================

 .. versionadded:: 2019.2.0

This module allows access to Redis  using an ``sdb://`` URI.

Like all SDB modules, the Redis module requires a configuration profile to
be configured in either the minion or master configuration file. This profile
requires very little. For example:

.. code-block:: yaml

    sdb_redis:
      driver: redis
      host: 127.0.0.1
      port: 6379
      password: pass
      db: 1

The ``driver`` refers to the Redis module, all other options are optional.
For option details see: https://redis-py.readthedocs.io/en/latest/.

"""

try:
    import redis

    HAS_REDIS = True
except ImportError:
    HAS_REDIS = False


__func_alias__ = {"set_": "set"}
__virtualname__ = "redis"


def __virtual__():
    """
    Module virtual name.
    """
    if not HAS_REDIS:
        return (False, "Please install python-redis to use this SDB module.")
    return __virtualname__


def set_(key, value, profile=None):
    """
    Set a value into the Redis SDB.
    """
    if not profile:
        return False
    redis_kwargs = profile.copy()
    redis_kwargs.pop("driver")
    redis_conn = redis.StrictRedis(**redis_kwargs)
    return redis_conn.set(key, value)


def get(key, profile=None):
    """
    Get a value from the Redis SDB.
    """
    if not profile:
        return False
    redis_kwargs = profile.copy()
    redis_kwargs.pop("driver")
    redis_conn = redis.StrictRedis(**redis_kwargs)
    return redis_conn.get(key)


def delete(key, profile=None):
    """
    Delete a key from the Redis SDB.
    """
    if not profile:
        return False
    redis_kwargs = profile.copy()
    redis_kwargs.pop("driver")
    redis_conn = redis.StrictRedis(**redis_kwargs)
    return redis_conn.delete(key)



import salt.modules.freezer as freezer
from tests.support.mock import patch


def test_compare(states, temp_salt_minion, tmp_path):
    """
    Test freezer.compare
    """
    # Default options
    opts = temp_salt_minion.config.copy()

    # Set cachedir
    cachedir = tmp_path / "__salt_test_freezer_cache_dir/minion"
    opts["cachedir"] = str(cachedir)

    # Freeze data setup
    freezedir = cachedir / "freezer"
    old_pkg_file = str(freezedir / "pre_install-pkgs.yml")
    old_rep_file = str(freezedir / "pre_install-reps.yml")
    new_pkg_file = str(freezedir / "post_install-pkgs.yml")
    new_rep_file = str(freezedir / "post_install-reps.yml")

    old_reps = {
        "http://deb.debian.org/debian": [
            {
                "file": "/etc/apt/sources.list",
                "comps": ["main"],
                "disabled": False,
                "dist": "buster",
                "type": "deb",
                "uri": "http://deb.debian.org/debian",
                "line": "deb http://deb.debian.org/debian buster main",
                "architectures": [],
            }
        ]
    }
    new_reps = {
        "http://deb.debian.org/debian": [
            {
                "file": "/etc/apt/sources.list",
                "comps": ["main"],
                "disabled": False,
                "dist": "buster",
                "type": "deb",
                "uri": "http://deb.debian.org/debian",
                "line": "deb http://deb.debian.org/debian buster main",
                "architectures": [],
            }
        ],
        "http://security.debian.org/debian-security": [
            {
                "file": "/etc/apt/sources.list",
                "comps": ["main"],
                "disabled": False,
                "dist": "buster/updates",
                "type": "deb",
                "uri": "http://security.debian.org/debian-security",
                "line": "deb http://security.debian.org/debian-security buster/updates main",
                "architectures": [],
            }
        ],
    }
    old_pkgs = {"adduser": "3.118"}
    new_pkgs = {"adduser": "3.118", "consul": "1.11.1"}

    # Mock up freeze files
    states.file.serialize(
        name=old_pkg_file, dataset=old_pkgs, serializer="json", makedirs=True
    )
    states.file.serialize(
        name=new_pkg_file, dataset=new_pkgs, serializer="json", makedirs=True
    )
    states.file.serialize(
        name=old_rep_file, dataset=old_reps, serializer="json", makedirs=True
    )
    states.file.serialize(
        name=new_rep_file, dataset=new_reps, serializer="json", makedirs=True
    )

    # Compare
    with patch("salt.modules.freezer.__opts__", opts, create=True):
        ret = freezer.compare(old="pre_install", new="post_install")

    assert ret["pkgs"]["new"]["consul"] == "1.11.1"
    assert "old" not in ret["pkgs"]
    assert len(ret["repos"]["new"]) == 1
    assert "old" not in ret["repos"]



"""
    :codeauthor: :email:`saltybob <bbaker@saltstack.com`
"""

from salt.client.ssh.wrapper import state


def test_parse_mods():
    """
    Test _parse_mods
    """
    expected = ["a", "b", "c", "d", "e", "f"]
    mods = "a,b, c,  d,e ,f  "

    actual = state._parse_mods(mods)
    assert expected == actual
